"""
Convenience wrappers for calling smart contracts via RPC.

The most specific wrappers in this module are for NEOs native contracts `NeoToken`, `GasToken`, `PolicyContract` and
 `RoleManagement`. The `ContractManagement` and `Ledger` contracts are not wrapped. See the FAQ for the reasons.

One step up are wrappers for contracts following the NEP-17 Token standard (`NEP17Contract`) and NEP-11 NFT standard
(`NEP11DivisibleContract` & `NEP11NonDivisibleContract`).
As last resort there is the `GenericContract` which can be used for calling arbitrary functions on arbitrary contracts.

Obtaining the execution results of the functions on the wrapped contracts is done through one of 3 methods on the
 ChainFacade class

1. `test_invoke()` - does not persist to chain. Costs no gas.
2. `invoke()` - does persist to chain. Requires signing and costs gas.
3. `estimate_gas()` - does not persist to chain. Costs no gas.

Example:
    facade = ChainFacade.node_provider_mainnet()

    neo = NeoToken()

    result = await facade.test_invoke(neo.candidates_registered())
"""

from __future__ import annotations
from typing import Callable, Any, TypeVar, Optional, cast, Generic, TypeAlias
from collections.abc import Sequence
import asyncio
from enum import IntEnum
from dataclasses import dataclass
from neo3.api import noderpc
from neo3.api.helpers import signing, txbuilder, unwrap
from neo3.network.payloads import verification
from neo3.wallet import utils as walletutils
from neo3.wallet.types import NeoAddress
from neo3.core import types, cryptography, utils as coreutils
from neo3 import vm
from neo3.contracts import contract, callflags, utils as contractutils, nef, manifest
from copy import deepcopy

# result stack index
ItemIndex: TypeAlias = int
ExecutionResultParser = Callable[[noderpc.ExecutionResult, ItemIndex], Any]

ReturnType = TypeVar("ReturnType")


class ContractMethodResult(Generic[ReturnType]):
    """
    A helper class around the `script` (VM opcodes) to be executed which allows to forward the annotated return type
    (`ContractMethodResult[T]`) of the annotated function `f` to a consumer function, without having to actually return
     an instance of `T` in the implementation of `f`.

     Example:
         def get_name() -> ContractMethodResult[str]:
            script = vm.ScriptBuilder().emit_contract_call(some_hash, "name")
            return ContractMethodResult(script, unwrap.as_str)

        def consumer(f: ContractMethodResult[ReturnType]) -> ReturnType:
            res = rpc.invoke_script(f.script)
            return f.execution_processor(res) # unwraps the result as a string

        x = consumer(get_name())
        type(x) # str
    """

    def __init__(
        self,
        script: bytes,
        execution_processor: Optional[ExecutionResultParser] = None,
        return_count: int = 1,
    ):
        """

        Args:
            script: VM opcodes to be executed.
            execution_processor: post processor function.
            return_count: number of items expected to be returned on the stack.
        """
        super(ContractMethodResult, self).__init__()
        self.script = script
        self.execution_processor = execution_processor
        self.return_count = return_count
        # TODO: add support for post processing notifications for functions that know will emit notifications
        self.notification_processor = None

    def __call__(self, *args, **kwargs):
        pass


_DEFAULT_MAINNET_RPC = "http://seed1.neo.org:10332"
_DEFAULT_TESTNET_RPC = "http://seed1t5.neo.org:20332"

SigningPair: TypeAlias = tuple[signing.SigningFunction, verification.Signer]


@dataclass
class InvokeReceipt(Generic[ReturnType]):
    """
    Transaction submission results.
    """

    #: Unique identifier.
    tx_hash: types.UInt256
    #: The block height/index the transaction is included in.
    included_in_block: int
    #: Number of blocks after accepting the transaction.
    confirmations: int
    #: The total gas cost for block inclusion and script execution.
    gas_consumed: int
    #: HALT = success, Others = failure.
    state: vm.VMState
    #: Virtual Machine exception.
    exception: Optional[str]
    #: Smart contract notifications.
    notifications: list[noderpc.Notification]
    #: Script excution result.
    result: ReturnType

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(tx_hash={str(self.tx_hash)}, included_in_block={self.included_in_block}, "
            f"confirmations={self.confirmations}, gas_consumed={self.gas_consumed}, state={self.state}, "
            f"exception={self.exception}, notifications={self.notifications}, result={self.result})"
        )


class ChainFacade:
    """
    The gateway to the network.

    Abstracts away the logic for talking to the data provider (NodeRPC only atm).
    """

    def __init__(
        self,
        rpc_host: str,
        receipt_retry_delay: Optional[float] = None,
        receipt_timeout: Optional[float] = None,
    ):
        """
        Args:
            rpc_host: Neo RPC node host address.
            receipt_retry_delay: time to wait in seconds between attempts to find the transaction on the chain.
            receipt_timeout: maximum time to wait in seconds to find the transaction on the chain.
        """
        self.rpc_host = rpc_host
        self._signing_func = None
        self.network = -1
        self.address_version = -1
        self.signers: list[verification.Signer] = []
        self._signing_funcs: list[signing.SigningFunction] = []
        self._receipt_retry_delay = receipt_retry_delay
        self._receipt_timeout = receipt_timeout

    async def test_invoke(
        self,
        f: ContractMethodResult[ReturnType],
        *,
        signers: Optional[Sequence[verification.Signer]] = None,
    ) -> ReturnType:
        """
        Call a contract method in read-only mode.
        This does not persist any state on the actual chain and therefore does not require signing or paying GAS.

        Args:
            f: function to call.
            signers: manually set the list of signers.

        See Also:
            `invoke()` - persists state.
        """
        if signers is None:
            signers = self.signers
        return await self._test_invoke(f, signers=signers)

    async def test_invoke_multi(
        self,
        f: list[ContractMethodResult],
        *,
        signers: Optional[Sequence[verification.Signer]] = None,
    ) -> Sequence:
        """
        Call all contract methods in one go (concurrently) and return the list of results.

        Args:
            f: list of functions to call.
            signers: manually set the list of signers.

        See Also:
            `invoke_multi()` - persists state.
        """
        if signers is None:
            signers = self.signers
        return await asyncio.gather(
            *map(lambda c: self.test_invoke(c, signers=signers), f)
        )

    async def test_invoke_raw(
        self,
        f: ContractMethodResult[ReturnType],
        *,
        signers: Optional[Sequence[verification.Signer]] = None,
    ) -> noderpc.ExecutionResult:
        """
        Call a contract method in read-only mode.
        This does not persist any state on the actual chain and therefore does not require signing or paying GAS.
        Does not post process the execution results.

        Args:
            f: function to call.
            signers: manually set the list of signers.

        See Also:
            `invoke_raw()` - persists state.
        """
        if signers is None:
            signers = self.signers
        return await self._test_invoke(f, signers=signers, return_raw=True)

    async def _test_invoke(
        self,
        f: ContractMethodResult[ReturnType],
        *,
        signers: Optional[Sequence[verification.Signer]] = None,
        return_raw: Optional[bool] = False,
    ):
        """
        Args:
            f:
            signers:
            return_raw: whether to post process the execution result or not.
        """
        async with noderpc.NeoRpcClient(self.rpc_host) as client:
            res = await client.invoke_script(f.script, signers)
            if f.execution_processor is None or return_raw:
                return res
            return f.execution_processor(res, 0)

    async def invoke(
        self,
        f: ContractMethodResult[ReturnType],
        *,
        signers: Optional[Sequence[SigningPair]] = None,
        network_fee: int = 0,
        system_fee: int = 0,
        append_network_fee: int = 0,
        append_system_fee: int = 0,
    ) -> InvokeReceipt[ReturnType]:
        """
        Call a contract method and persist results on the chain. Costs GAS.
        Waits for tx to be included in a block. Automatically post processes the execution results according to the
        post-processing function of `f` if present.

        Args:
            f: function to call.
            signers: manually set the list of signers.
            network_fee: manually set the network fee.
            system_fee: manually set the system fee.
            append_network_fee: increase the calculated network fee with this amount.
            append_system_fee: increase the calculated system fee with this amount.

        Returns:
            A transaction receipt. The `state` field of the receipt indicates if the transaction script executed
            successfully. The remaining fields provide additional information such as notifications that happened
            in the contract.

        See Also:
            `invoke_fast()` - does not wait for a receipt.
            `invoke_raw()` - does not wait for a transaction receipt, does not perform post-processing of the execution
                           results.
        """
        delay, timeout = await self._get_receipt_time_values()
        async with noderpc.NeoRpcClient(self.rpc_host) as client:
            tx_id = await self.invoke_fast(
                f,
                signers=signers,
                network_fee=network_fee,
                system_fee=system_fee,
                append_network_fee=append_network_fee,
                append_system_fee=append_system_fee,
            )
            receipt = await client.wait_for_transaction_receipt(
                tx_id, timeout=timeout, retry_delay=delay
            )
            if f.execution_processor is not None:
                result = f.execution_processor(receipt.execution, 0)
            else:
                result = receipt.execution
            return InvokeReceipt[ReturnType](
                receipt.tx_hash,
                receipt.included_in_block,
                receipt.confirmations,
                receipt.execution.gas_consumed,
                receipt.execution.state,
                receipt.execution.exception,
                receipt.execution.notifications,
                result,
            )

    async def invoke_fast(
        self,
        f: ContractMethodResult[ReturnType],
        *,
        signers: Optional[Sequence[SigningPair]] = None,
        network_fee: int = 0,
        system_fee: int = 0,
        append_network_fee: int = 0,
        append_system_fee: int = 0,
    ) -> types.UInt256:
        """
        Call a contract method and persist results on the chain. Costs GAS.

        Args:
            f: function to call.
            signers: manually set the list of signers.
            network_fee: manually set the network fee.
            system_fee: manually set the system fee.
            append_network_fee: increase the calculated network fee with this amount.
            append_system_fee: increase the calculated system fee with this amount.

        Returns:
            a transaction ID if accepted by the network. Acceptance is does not guarantee successful execution.
            Acceptance means there are no transaction format errors. Use `invoke()` to wait for a receipt. The `state`
            field of the receipt indicates if the transaction script executed successfully.

        See Also:
            `invoke()` - waits for a transaction receipt, performs post-processing of the execution results.
            `invoke_raw()` - does not wait for a transaction receipt, does not perform post-processing of the execution
                           results.
        """
        if network_fee > 0 and append_network_fee > 0:
            raise ValueError(
                "network_fee and append_network_fee are mutually exclusive"
            )

        if system_fee > 0 and append_system_fee > 0:
            raise ValueError("system_fee and append_system_fee are mutually exclusive")

        async with noderpc.NeoRpcClient(self.rpc_host) as client:
            builder = txbuilder.TxBuilder(client, f.script)
            await builder.init()

            if signers:
                for func, signer in signers:
                    builder.add_signer(func, signer)
            else:
                for func, signer in zip(self._signing_funcs, self.signers):
                    builder.add_signer(func, signer)

            await builder.set_valid_until_block()

            if system_fee > 0:
                builder.tx.system_fee = system_fee
            else:
                await builder.calculate_system_fee()
                builder.tx.system_fee += append_system_fee

            if network_fee > 0:
                builder.tx.network_fee = network_fee
            else:
                # calculate network fee has a chicken/egg problem for light wallet sdk's.
                # in order to calculate the right network fees (especially multisig) the witnesses have to be constructed
                # which is done in `build_and_sign()`.
                # at the same time `network_fee` is part of the signed data.
                # So here we call build_and_sign() just for creating the witnesses, then calculate the real fee (which will
                # reset the witnesses) and then at the end of the function we build_and_sign() again to have a valid
                # signature over the right network_fee

                # if there were no witnesses prior to signging, then we should restore that after the tmp signing
                reset_witnesses = len(builder.tx.witnesses) == 0
                builder.tx.network_fee = 999
                await builder.build_and_sign()
                await builder.calculate_network_fee()

                if reset_witnesses:
                    builder.tx.witnesses = []

                builder.tx.network_fee += append_network_fee

            tx = await builder.build_and_sign()
            return await client.send_transaction(tx)

    async def invoke_raw(
        self,
        f: ContractMethodResult[ReturnType],
        *,
        signers: Optional[Sequence[SigningPair]] = None,
        network_fee: int = 0,
        system_fee: int = 0,
        append_network_fee: int = 0,
        append_system_fee: int = 0,
    ) -> InvokeReceipt[noderpc.ExecutionResult]:
        """
        Call a contract method and persist results on the chain. Costs GAS.
        Waits for tx to be included in a block. Does not post processes the execution results.

        Args:
            f: function to call.
            signers: manually set the list of signers.
            network_fee: manually set the network fee.
            system_fee: manually set the system fee.
            append_network_fee: increase the calculated network fee with this amount.
            append_system_fee: increase the calculated system fee with this amount.

        Returns:
            A transaction receipt. The `state` field of the receipt indicates if the transaction script executed
            successfully. The remaining fields provide additional information such as notifications that happened
            in the contract.

        See Also:
            `invoke()` - waits for a transaction receipt, performs post-processing of the execution results
            `invoke_fast()` - does not wait for a receipt.
        """
        delay, timeout = await self._get_receipt_time_values()
        async with noderpc.NeoRpcClient(self.rpc_host) as client:
            tx_id = await self.invoke_fast(
                f,
                signers=signers,
                network_fee=network_fee,
                system_fee=system_fee,
                append_network_fee=append_network_fee,
                append_system_fee=append_system_fee,
            )
            receipt = await client.wait_for_transaction_receipt(
                tx_id, timeout=timeout, retry_delay=delay
            )

            return InvokeReceipt[noderpc.ExecutionResult](
                receipt.tx_hash,
                receipt.included_in_block,
                receipt.confirmations,
                receipt.execution.gas_consumed,
                receipt.execution.state,
                receipt.execution.exception,
                receipt.execution.notifications,
                receipt.execution.stack,
            )

    async def invoke_multi(
        self,
        f: list[ContractMethodResult],
        *,
        signers: Optional[Sequence[SigningPair]] = None,
        network_fee: int = 0,
        system_fee: int = 0,
        append_network_fee: int = 0,
        append_system_fee: int = 0,
        _post_processing: bool = True,
    ) -> Sequence:
        """
        Call all contract methods (concatenated) in one go and persist results on the chain. Costs GAS.
        Waits for tx to be included in a block. Automatically post processes the execution results according to the
        post-processing function of `f` if present.

        Args:
            f: list of functions to call.
            signers: manually set the list of signers.
            network_fee: manually set the network fee.
            system_fee: manually set the system fee.
            append_network_fee: increase the calculated network fee with this amount.
            append_system_fee: increase the calculated system fee with this amount.

        Returns:
            a list with the results of all exected functions.

        See Also:
            `test_invoke_multi` - free equivalent for read only operations or testing.
            `invoke_multi_fast` - does not wait for a receipt.
            `invoke_multi_raw` - does not wait for a transaction receipt, does not perform post-processing of the
                execution results.
        """
        tx_id = await self.invoke_multi_fast(
            f,
            signers=signers,
            network_fee=network_fee,
            system_fee=system_fee,
            append_network_fee=append_network_fee,
            append_system_fee=append_system_fee,
        )

        delay, timeout = await self._get_receipt_time_values()
        async with noderpc.NeoRpcClient(self.rpc_host) as client:
            receipt = await client.wait_for_transaction_receipt(
                tx_id, timeout=timeout, retry_delay=delay
            )

        results = []
        stack_offset = 0
        for call in f:
            res_cpy = deepcopy(receipt.execution)
            # adjust the stack so that it becomes transparent for the post-processing functions.
            res_cpy.stack = res_cpy.stack[
                stack_offset : stack_offset + call.return_count
            ]
            if call.execution_processor is None or not _post_processing:
                results.append(res_cpy)
            else:
                results.append(call.execution_processor(res_cpy, 0))
            stack_offset += call.return_count
        return results

    async def invoke_multi_fast(
        self,
        f: list[ContractMethodResult],
        *,
        signers: Optional[Sequence[SigningPair]] = None,
        network_fee: int = 0,
        system_fee: int = 0,
        append_network_fee: int = 0,
        append_system_fee: int = 0,
    ) -> types.UInt256:
        """
        Call all contract methods (concatenated) in one go and persist results on the chain. Costs GAS.
        Does not wait for tx to be included in a block. Automatically post processes the execution results according to
        the post-processing function of `f` if present.

        Args:
            f: list of functions to call.
            signers: manually set the list of signers.
            network_fee: manually set the network fee.
            system_fee: manually set the system fee.
            append_network_fee: increase the calculated network fee with this amount.
            append_system_fee: increase the calculated system fee with this amount.

        Returns:
            a transaction ID if accepted by the network. Acceptance is does not guarantee successful execution.
            Acceptance means there are no transaction format errors. Use `invoke()` to wait for a receipt. The `state`
            field of the receipt indicates if the transaction script executed successfully.


        See Also:
            `test_invoke_fast` - free equivalent for read only operations or testing.
            `invoke_multi` - waits for a transaction receipt, performs post-processing of the execution results.
            `invoke_multi_raw` - does not wait for a transaction receipt, does not perform post-processing of the
                execution results.
        """
        script = bytearray()
        for call in f:  # type: ContractMethodResult
            script.extend(call.script)

        wrapped: ContractMethodResult[None] = ContractMethodResult(script)
        tx_id = await self.invoke_fast(
            wrapped,
            signers=signers,
            network_fee=network_fee,
            system_fee=system_fee,
            append_network_fee=append_network_fee,
            append_system_fee=append_system_fee,
        )
        return tx_id

    async def invoke_multi_raw(
        self,
        f: list[ContractMethodResult],
        *,
        signers: Optional[Sequence[SigningPair]] = None,
        network_fee: int = 0,
        system_fee: int = 0,
        append_network_fee: int = 0,
        append_system_fee: int = 0,
    ) -> Sequence:
        """
        Call all contract methods (concatenated) in one go and persist results on the chain. Costs GAS.
        Do not wait for tx to be included in a block. Do not post process the execution results according to
        the post-processing function of `f` if present.

        Args:
            f: list of functions to call.
            signers: manually set the list of signers.
            network_fee: manually set the network fee.
            system_fee: manually set the system fee.
            append_network_fee: increase the calculated network fee with this amount.
            append_system_fee: increase the calculated system fee with this amount.

        Returns:
            a list with the results of all exected functions.

        See Also:
            `test_invoke_raw` - free equivalent for read only operations or testing.
            `invoke_multi` - waits for a transaction receipt, performs post-processing of the execution results.
            `invoke_multi_fast` - does not wait for a receipt.
        """
        return await self.invoke_multi(
            f,
            signers=signers,
            network_fee=network_fee,
            system_fee=system_fee,
            append_network_fee=append_network_fee,
            append_system_fee=append_system_fee,
            _post_processing=False,
        )

    async def estimate_gas(
        self,
        f: ContractMethodResult,
        *,
        signers: Optional[Sequence[verification.Signer]] = None,
    ) -> int:
        """
        Estimate the gas price for calling the contract method.
        """
        async with noderpc.NeoRpcClient(self.rpc_host) as client:
            res = await client.invoke_script(f.script, signers)
            return res.gas_consumed

    async def _get_receipt_time_values(self) -> tuple[float, float]:
        if self._receipt_retry_delay is None or self._receipt_timeout is None:
            async with noderpc.NeoRpcClient(self.rpc_host) as client:
                result = await client.get_version()
                # 5 seems like a reasonable divider where on mainnet (with 15s blocks) at worst case
                # the RPC server is queried 5 times.
                delay = self._receipt_retry_delay = (
                    result.protocol.ms_per_block / 1000
                ) / 5
                timeout = self._receipt_timeout = (
                    (result.protocol.ms_per_block / 1000) * 2
                ) + self._receipt_retry_delay
                return delay, timeout
        else:
            return self._receipt_retry_delay, self._receipt_timeout

    def add_signer(self, func: signing.SigningFunction, signer: verification.Signer):
        """
        Add a `Signer` which will automatically be included when the various invoke functions.
        """
        self._signing_funcs.append(func)
        self.signers.append(signer)

    @classmethod
    def node_provider_mainnet(cls):
        """
        Create a facade pre-configured to N3 MainNet.
        """
        return cls(_DEFAULT_MAINNET_RPC)

    @classmethod
    def node_provider_testnet(cls):
        """
        Create a facade pre-configured to N3 TestNet.
        """
        return cls(_DEFAULT_TESTNET_RPC)


class GenericContract:
    """
    Generic class to call arbitrary methods on a smart contract.
    """

    def __init__(self, contract_hash: types.UInt160):
        self.hash = contract_hash

    def call_function(
        self,
        name,
        args: Optional[noderpc.ContractParameter] = None,
    ) -> ContractMethodResult[noderpc.ExecutionResultResponse]:
        """
        Call a method on the contract.

        Args:
            name: the method name to call as defined in the manifest.
            args: optional list of arguments the function expects.
        """
        if args is None:
            script = vm.ScriptBuilder().emit_contract_call(self.hash, name).to_array()
        else:
            sb = vm.ScriptBuilder()
            sb.emit_contract_call_with_args(self.hash, name, args)
            script = sb.to_array()
        return ContractMethodResult(script)

    def update(
        self,
        update_method: str = "update",
        nef: Optional[nef.NEF] = None,
        manifest: Optional[manifest.ContractManifest] = None,
        data: Optional[noderpc.ContractParameter] = None,
    ) -> ContractMethodResult[None]:
        """
        Update this contract on chain with a new manifest and/or contract (NEF).

        Assumes the update method on the contract uses the standard arguments; nef, manifest and (optional) data.
        If it uses custom arguments use `call_function` instead.

        Args:
            update_method: override with name of the update function on the contract.
            nef: compiled contract.
            manifest: contract manifest.
            data: data that is passed to the `_deploy` method of the smart contract (if the method exists).
        """
        if nef is None and manifest is None:
            raise ValueError("NEF and manifest are both None. Nothing to update")

        m: Optional[str] = None
        if manifest is not None:
            # convert to the right format as expected by the management contract
            m = str(manifest)

        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(self.hash, update_method, [nef, m, data])

        return ContractMethodResult(sb.to_array(), unwrap.as_none)

    def destroy(self, destroy_method: str = "destroy") -> ContractMethodResult[None]:
        """
        Destroy the contract on chain.

        This will permanently block the contract hash on the network.

        Args:
            destroy_method: override with name of the destroy function on the contract. Default signature is `destroy()`.
        """
        sb = vm.ScriptBuilder().emit_contract_call(self.hash, destroy_method)
        return ContractMethodResult(sb.to_array(), unwrap.as_none)

    @staticmethod
    def deploy(
        nef: nef.NEF,
        manifest: manifest.ContractManifest,
        data: Optional[noderpc.ContractParameter] = None,
    ) -> ContractMethodResult[types.UInt160]:
        """
        Deploy a smart contract to the chain.

        Args:
            nef: compiled contract.
            manifest: contract manifest file.
            data: data that is passed to the `_deploy` method of the smart contract (if the method exists).

        Returns:
            contract hash.
        """
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(
            contract.CONTRACT_HASHES.MANAGEMENT, "deploy", [nef, str(manifest), data]
        )

        def process(res, _):
            arr = unwrap.as_list(res)
            return arr[2].as_uint160()

        return ContractMethodResult(sb.to_array(), process)


class _TokenContract(GenericContract):
    """
    Base class for Fungible and Non-Fungible tokens
    """

    def symbol(self) -> ContractMethodResult[str]:
        """
        User-friendly name of the token.
        """
        script = vm.ScriptBuilder().emit_contract_call(self.hash, "symbol").to_array()
        return ContractMethodResult(script, unwrap.as_str)

    def decimals(self) -> ContractMethodResult[int]:
        """
        Get the amount of decimals.

        Use this with the result of balance_of to display the correct user representation.
        """
        script = vm.ScriptBuilder().emit_contract_call(self.hash, "decimals").to_array()
        return ContractMethodResult(script, unwrap.as_int)

    def total_supply(self) -> ContractMethodResult[int]:
        """
        Get the total token supply in the NEO system.
        """
        script = (
            vm.ScriptBuilder().emit_contract_call(self.hash, "totalSupply").to_array()
        )
        return ContractMethodResult(script, unwrap.as_int)


class NEP17Contract(_TokenContract):
    """
    Base class for calling NEP-17 compliant smart contracts.
    """

    def balance_of(
        self, account: types.UInt160 | NeoAddress
    ) -> ContractMethodResult[int]:
        """
        Get the balance for the given account.

        Note:
            The returned value does not take the token decimals into account. e.g. for the GAS token you want to
        divide the result by 10**8 (as the Gas token has 8 decimals).

        Raises:
            ValueError: if `account` is an invalid NeoAddress format.
        """
        account = _check_address_and_convert(account)
        script = (
            vm.ScriptBuilder()
            .emit_contract_call_with_args(self.hash, "balanceOf", [account])
            .to_array()
        )
        return ContractMethodResult(script, unwrap.as_int)

    def balance_of_friendly(
        self, account: types.UInt160 | NeoAddress
    ) -> ContractMethodResult[float]:
        """
        Get the balance for the given account while taking the token decimals into account.

        Uses the token decimals to convert to the user end representation.

        Raises:
            ValueError: if `account` is an invalid NeoAddress format.
        """
        account = _check_address_and_convert(account)

        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(self.hash, "balanceOf", [account])
        sb.emit_contract_call(self.hash, "decimals")

        def process(res: noderpc.ExecutionResult, _: int = 0) -> float:
            unwrap.check_state_ok(res)
            balance = unwrap.as_int(res, 0)
            decimals = unwrap.as_int(res, 1)
            if balance == 0:
                return 0.0
            else:
                return balance / (10**decimals)

        return ContractMethodResult(sb.to_array(), process)

    def transfer(
        self,
        source: types.UInt160 | NeoAddress,
        destination: types.UInt160 | NeoAddress,
        amount: int,
        data: Optional[noderpc.ContractParameter] = None,
    ) -> ContractMethodResult[bool]:
        """
        Transfer `amount` of tokens from `source` account to `destination` account.
        Forward `data` to `onNEP17Payment` handler if applicable.

        For this to pass while using `test_invoke()`, make sure to add a Signer with a script hash equal to the source
        account. i.e.

            source = <source_script_hash>
            signer = verification.Signer(source, payloads.WitnessScope.CALLED_BY_ENTRY)
            await facade.test_invoke(token.transfer(source, destination, 10), signers=[signer]))

        Raises:
            ValueError: if `source` or `destination` is an invalid NeoAddress format

        Returns:
            The return value after invoking with `invoke()` or `test_invoke()` is
                `True` if the token transferred successful.
                `False` otherwise.
        """
        source = _check_address_and_convert(source)
        destination = _check_address_and_convert(destination)

        sb = vm.ScriptBuilder().emit_contract_call_with_args(
            self.hash, "transfer", [source, destination, amount, data]
        )
        return ContractMethodResult(sb.to_array(), unwrap.as_bool)

    def transfer_friendly(
        self,
        source: types.UInt160 | NeoAddress,
        destination: types.UInt160 | NeoAddress,
        amount,
        data: Optional[noderpc.ContractParameter] = None,
    ):
        """
        Transfer `amount` of tokens from `source` account to `destination` account.

        Same as `transfer` but does not require manually convert amount if the token uses decimals.

        Returns:
            The return value after invoking with `invoke()` or `test_invoke()` is
                `True` if the token transferred successful.
                `False` otherwise.

        Raises:
            ValueError: if `source` or `destination` is an invalid NeoAddress format
        """
        source = _check_address_and_convert(source)
        destination = _check_address_and_convert(destination)
        sb = vm.ScriptBuilder()
        sb.emit_push(data)
        sb.emit_push(10)  # multiplier base
        sb.emit_contract_call(self.hash, "decimals")
        sb.emit(vm.OpCode.POW)
        sb.emit_push(amount)
        sb.emit(vm.OpCode.MUL)
        sb.emit_push(destination)
        sb.emit_push(source)
        sb.emit_push(4)
        sb.emit(vm.OpCode.PACK)
        sb.emit_push(callflags.CallFlags.ALL)
        sb.emit_push("transfer")
        sb.emit_push(self.hash)
        sb.emit_syscall(vm.Syscalls.SYSTEM_CONTRACT_CALL)

        return ContractMethodResult(sb.to_array(), unwrap.as_bool)

    def transfer_multi(
        self,
        source: types.UInt160 | NeoAddress,
        destinations: Sequence[types.UInt160 | NeoAddress],
        amount: int,
        data: Optional[noderpc.ContractParameter] = None,
        abort_on_failure: bool = False,
    ) -> ContractMethodResult[bool]:
        """
        Transfer `amount` of tokens from `source` to each account in `destinations`.

        Args:
            source: account to take funds from.
            destinations: accounts to send funds to.
            amount: how much to transfer.
            data: forward to `onNEP17Payment` handler if applicable.
            abort_on_failure: if True aborts the whole transaction if any of the transfers fails.

        Raises:
            ValueError: if any of the destinations is supplied as NeoAddress but is an invalid address format.

        Returns:
            The return value after invoking with `invoke()` or `test_invoke()` is
                `True` if all transfers are successful.
                `False` otherwise.
        """
        sb = vm.ScriptBuilder()
        source = _check_address_and_convert(source)
        for d in destinations:
            d = _check_address_and_convert(d)
            sb.emit_contract_call_with_args(
                self.hash, "transfer", [source, d, amount, data]
            )
            if abort_on_failure:
                sb.emit(vm.OpCode.ASSERT)

        # when abort_on_failure is used the result of the `transfer()` call is consumed by the ASSERT opcode
        # and the `stack` will be empty. Therefore, we only check for the VM state
        def process_with_assert(res: noderpc.ExecutionResult, _: int = 0) -> bool:
            unwrap.check_state_ok(res)
            return True

        # when abort_on_failure is not used we iterate over all transfer() results
        def process(res: noderpc.ExecutionResult, _: int = 0) -> bool:
            unwrap.check_state_ok(res)
            for si in res.stack:
                if si.as_bool() is False:
                    return False
            return True

        if abort_on_failure:
            return ContractMethodResult(sb.to_array(), process_with_assert)
        else:
            return ContractMethodResult(sb.to_array(), process)


class GasToken(NEP17Contract):
    """
    Wrapped GAS token contract.
    """

    def __init__(self):
        super(GasToken, self).__init__(contract.CONTRACT_HASHES.GAS_TOKEN)


class Candidate:
    """
    Container for holding consensus candidate voting results.
    """

    def __init__(self, public_key: cryptography.ECPoint, votes: int):
        #: public key of the candidate.
        self.public_key = public_key
        #: number of votes the candidate has.
        self.votes = votes
        shash = coreutils.to_script_hash(
            contractutils.create_signature_redeemscript(self.public_key)
        )
        #: NEO address of the candidate.
        self.address = walletutils.script_hash_to_address(shash)


class NeoToken(NEP17Contract):
    """
    Wrapped NEO token contract.
    """

    def __init__(self):
        super(NeoToken, self).__init__(contract.CONTRACT_HASHES.NEO_TOKEN)

    def get_gas_per_block(self) -> ContractMethodResult[int]:
        """
        Get the amount of GAS generated in each block.
        """
        script = (
            vm.ScriptBuilder()
            .emit_contract_call(self.hash, "getGasPerBlock")
            .to_array()
        )
        return ContractMethodResult(script, unwrap.as_int)

    def get_unclaimed_gas(
        self, account: types.UInt160 | NeoAddress, end: Optional[int] = None
    ) -> ContractMethodResult[int]:
        """
        Get the amount of unclaimed GAS.

        Args:
            account: for whom.
            end: up to which block height to calculate the GAS bonus. Omit to calculate to the current chain height.

        Raises:
            ValueError: if `account` is an invalid NeoAddress format.
        """
        if not isinstance(account, types.UInt160):
            walletutils.validate_address(account)
            account = walletutils.address_to_script_hash(account)

        sb = vm.ScriptBuilder()
        if end is not None:
            sb.emit_contract_call_with_args(
                self.hash, "unclaimedGas", [account, end]
            ).to_array()
            return ContractMethodResult(sb.to_array(), unwrap.as_int)
        else:
            sb.emit_contract_call(contract.CONTRACT_HASHES.LEDGER, "currentIndex")
            # first argument to "unclaimedGas" (second is already on the stack by the "currentIndex" call
            sb.emit_push(account)
            # length of arguments
            sb.emit_push(2)
            sb.emit(vm.OpCode.PACK)
            sb.emit_push(callflags.CallFlags.ALL)
            sb.emit_push("unclaimedGas")
            sb.emit_push(self.hash)
            sb.emit_syscall(vm.Syscalls.SYSTEM_CONTRACT_CALL)
            return ContractMethodResult(sb.to_array(), unwrap.as_int)

    def candidate_registration_price(self) -> ContractMethodResult[int]:
        """
        Get the amount of GAS to pay to register as consensus candidate.
        """
        script = (
            vm.ScriptBuilder()
            .emit_contract_call(self.hash, "getRegisterPrice")
            .to_array()
        )
        return ContractMethodResult(script, unwrap.as_int)

    def candidate_register(
        self, public_key: cryptography.ECPoint
    ) -> ContractMethodResult[bool]:
        """
        Register as a consensus candidate.

        See Also:
            Account.public_key

        Returns:
            `True` if successful. `False` otherwise.
        """
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(
            self.hash, "registerCandidate", [public_key]
        ).to_array()
        return ContractMethodResult(sb.to_array(), unwrap.as_bool)

    def candidate_unregister(
        self, public_key: cryptography.ECPoint
    ) -> ContractMethodResult[bool]:
        """
        Unregister as a consensus candidate.

        See Also:
            Account.public_key

        Returns:
            `True` if successful. `False` otherwise.
        """
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(
            self.hash, "registerCandidate", [public_key]
        ).to_array()
        return ContractMethodResult(sb.to_array(), unwrap.as_bool)

    def candidate_vote(
        self, voter: types.UInt160 | NeoAddress, candidate: cryptography.ECPoint
    ) -> ContractMethodResult[bool]:
        """
        Cast a vote for `candidate` to become a consensus node.

        Args:
            voter: the account to vote from.
            candidate: who to vote on.

        Raises:
            ValueError: if `voter` is an invalid NeoAddress format.

        Returns:
            `True` if vote cast successful. `False` otherwise.
        """
        voter = _check_address_and_convert(voter)
        script = (
            vm.ScriptBuilder()
            .emit_contract_call_with_args(self.hash, "vote", [voter, candidate])
            .to_array()
        )
        return ContractMethodResult(script, unwrap.as_bool)

    def candidate_votes(
        self, candidate: cryptography.ECPoint
    ) -> ContractMethodResult[int]:
        """
        Get the total vote count for `candidate`.
        """
        script = (
            vm.ScriptBuilder()
            .emit_contract_call_with_args(self.hash, "getCandidateVote", [candidate])
            .to_array()
        )
        return ContractMethodResult(script, unwrap.as_int)

    def candidates_registered(self) -> ContractMethodResult[list[Candidate]]:
        """
        Get the first 256 registered candidates.
        """
        script = (
            vm.ScriptBuilder().emit_contract_call(self.hash, "getCandidates").to_array()
        )

        def process(res: noderpc.ExecutionResult, _: int = 1) -> list[Candidate]:
            raw_results: list[noderpc.StackItem] = unwrap.as_list(res)
            result = []

            for si in raw_results:
                if si.type != noderpc.StackItemType.STRUCT:
                    continue
                v = cast(list[noderpc.StackItem], si.value)
                result.append(Candidate(v[0].as_public_key(), v[1].as_int()))
            return result

        return ContractMethodResult(script, process)


class _NEP11Contract(_TokenContract):
    """
    Base class for calling NEP-11 compliant smart contracts.

    NFTs can be divisible or non-divisible which is determined by the value of `decimals()`.

    Note:
        The following 2 common methods defined in the NEP-11 standard have different names to improve discoverability
         NEP-11 standard    This wrapper
      1. balanceOf       -> total_owned_by
      2. tokensOf        -> token_ids_owned_by
    """

    def decimals(self) -> ContractMethodResult[int]:
        """
        Get the amount of decimals.

        A zero return value indicates a non-divisible NFT.
        A bigger than zero return value indicates a divisible NFTs.
        """
        return super(_NEP11Contract, self).decimals()

    def total_owned_by(
        self, owner: types.UInt160 | NeoAddress
    ) -> ContractMethodResult[int]:
        """
        Get the total amount of NFTs owned for the given account.

        Raises:
            ValueError: if `owner` is an invalid NeoAddress format.
        """
        owner = _check_address_and_convert(owner)
        script = (
            vm.ScriptBuilder()
            .emit_contract_call_with_args(self.hash, "balanceOf", [owner])
            .to_array()
        )
        return ContractMethodResult(script, unwrap.as_int)

    def token_ids_owned_by(
        self, owner: types.UInt160 | NeoAddress
    ) -> ContractMethodResult[list[bytes]]:
        """
        Get an iterator containing all token ids owned by the specified address.

        Raises:
            ValueError: if `owner` is an invalid NeoAddress format.
        """
        owner = _check_address_and_convert(owner)
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args_and_unwrap_iterator(
            self.hash, "tokensOf", [owner]
        )

        def process(res: noderpc.ExecutionResult, _: int = 0) -> list[bytes]:
            raw_results: list[noderpc.StackItem] = unwrap.as_list(res)
            return [si.value for si in raw_results]

        return ContractMethodResult(sb.to_array(), process)

    def tokens(
        self, limit: int = 2000, start_index: int = 0
    ) -> ContractMethodResult[list[bytes]]:
        """
        Get all tokens minted by the contract.

        limit: the maximum tokens to return.
        start_index: where to start collecting results from (see Note1)

        Note1:
            There is a limit on the returned items in the virtual machine (default: 2048) to avoid too
            much compute being used. The limit here is set slightly lower on purpose to allow other necessary items in the
            VM. If the contract returns more than 2000 items you need to call the method again providing the start_index
            parameter. The `tokens_count` method can be used to get the size of the iterator.

        Note2:
            This is an optional method in the NEP-11 standard and can thus fail.
        """

        def process(res: noderpc.ExecutionResult, _: int = 0) -> list[bytes]:
            raw_results: list[noderpc.StackItem] = unwrap.as_list(res)
            return [si.value for si in raw_results]

        sb = vm.ScriptBuilder().emit_contract_call_and_unwrap_iterator(
            self.hash, "tokens", unwrap_limit=limit, start_index=start_index
        )
        return ContractMethodResult(sb.to_array(), process)

    def tokens_count(self):
        """
        Count all tokens minted by the contract

        Note:
            This depends on the optional `tokens` method and can thus fail.
        """
        sb = vm.ScriptBuilder().emit_contract_call_and_count_iterator(
            self.hash, "tokens"
        )
        return ContractMethodResult(sb.to_array(), unwrap.as_int)

    def properties(self, token_id: bytes) -> ContractMethodResult[dict]:
        """
        Get all properties for the given NFT.

        Note:
            This is an optional method and may not exist on the contract.
        """
        sb = vm.ScriptBuilder().emit_contract_call_with_args(
            self.hash, "properties", [token_id]
        )
        return ContractMethodResult(sb.to_array(), unwrap.as_dict)


class NEP11DivisibleContract(_NEP11Contract):
    """
    Base class for divisible NFTs.

    The NEP-11 `ownerOf` method is named `owners_of` in this wrapper.
    """

    def transfer(
        self,
        source: types.UInt160 | NeoAddress,
        destination: types.UInt160 | NeoAddress,
        amount: int,
        token_id: bytes,
        data: Optional[noderpc.ContractParameter] = None,
    ) -> ContractMethodResult[bool]:
        """
        Transfer `amount` of `token_id` from `source` account to `destination` account.

        For this to pass while using `test_invoke()`, make sure to add a Signer with a script hash equal to the source
        account. i.e.

            source = <source_script_hash>
            signer = verification.Signer(source, payloads.WitnessScope.CALLED_BY_ENTRY)
            await facade.test_invoke(token.transfer(source, destination, 10), signers=[signer]))

        Raises:
            ValueError: if `source` or `destination` is an invalid NeoAddress format.

        Returns:
            True if token fractions transferred successful. False otherwise.
        """
        source = _check_address_and_convert(source)
        destination = _check_address_and_convert(destination)
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(
            self.hash, "transfer", [source, destination, amount, token_id, data]
        )
        return ContractMethodResult(sb.to_array(), unwrap.as_bool)

    def owners_of(self, token_id: bytes) -> ContractMethodResult[list[types.UInt160]]:
        """
        Get a list of account script hashes that own a part of `token_id`.
        """
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args_and_unwrap_iterator(
            self.hash, "ownerOf", [token_id]
        )

        def process(res: noderpc.ExecutionResult, _: int = 0) -> list[types.UInt160]:
            raw_results: list[noderpc.StackItem] = unwrap.as_list(res)
            return [si.as_uint160() for si in raw_results]

        return ContractMethodResult(sb.to_array(), process)

    def balance_of(
        self, owner: types.UInt160 | NeoAddress, token_id: bytes
    ) -> ContractMethodResult[int]:
        """
        Get the token balance for the given owner.

        Raises:
            ValueError: if `owner` is an invalid NeoAddress format.
        """
        owner = _check_address_and_convert(owner)
        sb = vm.ScriptBuilder().emit_contract_call_with_args(
            self.hash, "balanceOf", [owner, token_id]
        )
        return ContractMethodResult(sb.to_array(), unwrap.as_int)

    def balance_of_friendly(
        self, owner: types.UInt160 | NeoAddress, token_id: bytes
    ) -> ContractMethodResult[float]:
        """
        Get the balance for the given account and convert the result into the user representation.

        Uses the token decimals to convert to the user end representation.

        Raises:
            ValueError: if `owner` is an invalid NeoAddress format.
        """
        owner = _check_address_and_convert(owner)
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(self.hash, "balanceOf", [owner, token_id])
        sb.emit_contract_call(self.hash, "decimals")

        def process(res: noderpc.ExecutionResult, _: int = 0) -> float:
            unwrap.check_state_ok(res)
            balance = unwrap.as_int(res, 0)
            decimals = unwrap.as_int(res, 1)
            if balance == 0:
                return 0.0
            else:
                return balance / (10**decimals)

        return ContractMethodResult(sb.to_array(), process)


class NEP11NonDivisibleContract(_NEP11Contract):
    """Base class for non-divisible NFTs."""

    def transfer(
        self,
        destination: types.UInt160 | NeoAddress,
        token_id: bytes,
        data: Optional[noderpc.ContractParameter] = None,
    ) -> ContractMethodResult[bool]:
        """
        Transfer `token_id` to `destination` account.
        The source account will be the account that pays for the fees (a.k.a. the transaction.sender).

        For this to pass while using `test_invoke()`, make sure to add a Signer with a script hash equal to the source
        account. i.e.

            signer = verification.Signer(source_account, payloads.WitnessScope.CALLED_BY_ENTRY)
            await facade.test_invoke(token.transfer(destination, token_id, 10), signers=[signer]))

        Raises:
            ValueError: if `destination` is an invalid NeoAddress format.

        Returns:
            The return value after invoking with `invoke()` or `test_invoke()` is
                `True` if the token transferred successful.
                `False` otherwise.
        """
        destination = _check_address_and_convert(destination)
        sb = vm.ScriptBuilder().emit_contract_call_with_args(
            self.hash, "transfer", [destination, token_id, data]
        )
        return ContractMethodResult(sb.to_array(), unwrap.as_bool)

    def transfer_multi(
        self,
        destinations: Sequence[types.UInt160 | NeoAddress],
        token_ids: list[bytes],
        data: Optional[noderpc.ContractParameter] = None,
        abort_on_failure: bool = False,
    ) -> ContractMethodResult[bool]:
        """
        Transfer multiple token_ids to multiple destinations. Destination and token_ids are paired in order.

        Args:
            destinations: accounts to send tokens to.
            token_ids: list of token ids.
            data: forward to `onNEP17Payment` handler if applicable.
            abort_on_failure: if `True` aborts the whole transaction if any of the transfers fails.

        Raises:
            ValueError: if any of the destinations is supplied as NeoAddress but is an invalid address format.

        Returns:
            The return value after invoking with `invoke()` or `test_invoke()` is
                `True` if all transfers are successful.
                `False` otherwise.
        """
        sb = vm.ScriptBuilder()
        for d, t in zip(destinations, token_ids):
            d = _check_address_and_convert(d)
            sb.emit_contract_call_with_args(self.hash, "transfer", [d, t, data])
            if abort_on_failure:
                sb.emit(vm.OpCode.ASSERT)

        # when abort_on_failure is used the result of the `transfer()` call is consumed by the ASSERT opcode
        # and the `stack` will be empty. Therefore, we only check for the VM state
        def process_with_assert(res: noderpc.ExecutionResult, _: int = 0) -> bool:
            unwrap.check_state_ok(res)
            return True

        # when abort_on_failure is not used we iterate over all transfer() results
        def process(res: noderpc.ExecutionResult, _: int = 0) -> bool:
            unwrap.check_state_ok(res)
            for si in res.stack:
                if si.as_bool() is False:
                    return False
            return True

        if abort_on_failure:
            return ContractMethodResult(sb.to_array(), process_with_assert)
        else:
            return ContractMethodResult(sb.to_array(), process)

    def owner_of(self, token_id: bytes) -> ContractMethodResult[types.UInt160]:
        """
        Get the owner of the given token.
        """
        sb = vm.ScriptBuilder().emit_contract_call_with_args(
            self.hash, "ownerOf", [token_id]
        )
        return ContractMethodResult(sb.to_array(), unwrap.as_uint160)


def _check_address_and_convert(value: types.UInt160 | NeoAddress) -> types.UInt160:
    if isinstance(value, types.UInt160):
        return value
    if not isinstance(value, str):
        raise ValueError(
            f"Input is of type {type(value)} expected UInt160 or NeoAddress(str)"
        )
    walletutils.validate_address(value)
    return walletutils.address_to_script_hash(value)


class PolicyContract(GenericContract):
    """
    Wrapper around the native contract that manages system policies.

    Note:
        Functions that only consensus committee members can use are not implemented.
    """

    def __init__(self):
        super(PolicyContract, self).__init__(contract.CONTRACT_HASHES.POLICY)

    def fee_per_byte(self) -> ContractMethodResult[int]:
        """
        The fee per transaction byte.
        """
        sb = vm.ScriptBuilder().emit_contract_call(self.hash, "getFeePerByte")
        return ContractMethodResult(sb.to_array(), unwrap.as_int)

    def exec_fee_factor(self) -> ContractMethodResult[int]:
        """
        The system fee multiplier for transactions.
        """
        sb = vm.ScriptBuilder().emit_contract_call(self.hash, "getExecFeeFactor")
        return ContractMethodResult(sb.to_array(), unwrap.as_int)

    def storage_price(self) -> ContractMethodResult[int]:
        """
        The price per byte of smart contract storage.
        """
        sb = vm.ScriptBuilder().emit_contract_call(self.hash, "getStoragePrice")
        return ContractMethodResult(sb.to_array(), unwrap.as_int)

    def is_blocked(self, script_hash: types.UInt160) -> ContractMethodResult[bool]:
        """
        Check if an account or contract is blocked on the network.
        """
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(self.hash, "isBlocked", [script_hash])
        return ContractMethodResult(sb.to_array(), unwrap.as_bool)


class DesignateRole(IntEnum):
    """
    Special role that can be assigned to nodes in the network by the consensus committee.
    """

    STATE_VALIDATOR = 4
    ORACLE = 8
    NEO_FS_ALPHABET_NODE = 16


class RoleContract(GenericContract):
    """
    Wrapper around the native Role management contract.
    """

    def __init__(self):
        super(RoleContract, self).__init__(contract.CONTRACT_HASHES.ROLE_MANAGEMENT)

    def get_designated_by_role(
        self, role: DesignateRole, block_index: int
    ) -> ContractMethodResult[list[cryptography.ECPoint]]:
        """
        Gets the public keys registered for a given role at a given height.
        """
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(
            self.hash, "getDesignatedByRole", [role, block_index]
        )

        def process(res: noderpc.ExecutionResult, _: int):
            arr = unwrap.as_list(res)
            return list(map(lambda x: x.as_public_key(), arr))

        return ContractMethodResult(sb.to_array(), process)



"""
P2P network classes. Holds `Block` and `Transaction` payloads (among others), a network node and helper classes for
 syncing the chain.
"""
import sys
import asyncio

if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())



from __future__ import annotations
from jsonschema import validate  # type: ignore
from neo3.core import interfaces


class ScryptParameters(interfaces.IJson):
    json_schema = {
        "type": "object",
        "properties": {
            "n": {"type": "integer"},
            "r": {"type": "integer"},
            "p": {"type": "integer"},
        },
        "required": ["n", "r", "p"],
    }

    def __init__(self, n: int = 16384, r: int = 8, p: int = 8):
        self.n = n
        self.r = r
        self.p = p

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        return {"n": self.n, "r": self.r, "p": self.p}

    @classmethod
    def from_json(cls, json: dict) -> ScryptParameters:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
        """
        validate(json, schema=cls.json_schema)

        return cls(n=json["n"], r=json["r"], p=json["p"])



"""
Helper functions to easily fetch native values from the `ResultStack` returned as response by various RPC methods
such as `invoke_function()`, `invoke_script()`, `get_application_log_transaction()` and `get_application_log_block()`.

Includes sanity checking.
"""
from __future__ import annotations
from neo3.api import noderpc
from neo3 import vm
from neo3.core import types, cryptography


def check_state_ok(res: noderpc.ExecutionResult):
    """
    Check if the execution of the transaction finished in a success state.

    Raises:
        ValueError: if the VM state is not HALT.
    """
    if vm.VMState.from_string(res.state) != vm.VMState.HALT:
        raise ValueError(
            f"Transaction execution failed with state {res.state} and err: {res.exception}"
        )


def as_bool(res: noderpc.ExecutionResult, idx: int = 0) -> bool:
    """
    Convert the stack item at `idx` to a `bool`.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to a bool.
    """
    return item(res, idx).as_bool()


def as_str(res: noderpc.ExecutionResult, idx: int = 0) -> str:
    """
    Convert the stack item at `idx` to a `str`.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to a `str`.
    """
    return item(res, idx).as_str()


def as_int(res: noderpc.ExecutionResult, idx: int = 0) -> int:
    """
    Convert the stack item at `idx` to an `int`.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to an int.
    """
    return item(res, idx).as_int()


def as_uint160(res: noderpc.ExecutionResult, idx: int = 0) -> types.UInt160:
    """
    Convert the stack item at `idx` to an `UInt160`.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to an UInt160.
    """
    return item(res, idx).as_uint160()


def as_uint256(res: noderpc.ExecutionResult, idx: int = 0) -> types.UInt256:
    """
    Convert the stack item at `idx` to an `UInt256`.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to an UInt256.
    """
    return item(res, idx).as_uint256()


def as_address(res: noderpc.ExecutionResult, idx: int = 0) -> str:
    """
    Convert the stack item at `idx` to a NEO3 address.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to a NEO3 address.
    """
    return item(res, idx).as_address()


def as_public_key(res: noderpc.ExecutionResult, idx: int = 0) -> cryptography.ECPoint:
    """
    Convert the stack item at `idx` to a public key.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to a bool.
        ECCException: if the resulting key is not valid on the SECP256R1 curve.
    """
    return item(res, idx).as_public_key()


def as_list(res: noderpc.ExecutionResult, idx: int = 0) -> list[noderpc.StackItem]:
    """
    Convert the stack item at `idx` to a `list`.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to a list.
    """
    return item(res, idx).as_list()


def as_dict(res: noderpc.ExecutionResult, idx: int = 0) -> dict:
    """
    Convert the stack item at `idx` to a dictionary.

    Args:
        res: execution result.
        idx: idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to a dict.


    Warning:
        Accepted key types on the Virtual Machine side are `int`, `str` and `bytes`.
        However, when data is returned there is no way to differentiate between
        the key being of type `str` or `bytes` as the RPC node will encode both as a `ByteString`.
        The dictionary returned will return such types as `bytes` and it is the user responsibility to decode
        them to a `str` if needed.
    """
    return item(res, idx).as_dict()


def as_none(res: noderpc.ExecutionResult, idx: int = 0) -> None:
    """
    Convert the stack item at `idx` to `None`.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value is not `None`.
    """
    return item(res, idx).as_none()


def as_bytes(res: noderpc.ExecutionResult, idx: int = 0) -> bytes:
    """
    Convert the stack item at `idx` to `bytes`.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.

    Raises:
        ValueError: if the index is out of range, or the value cannot be converted to bytes.

    """
    return item(res, idx).as_bytes()


def item(res: noderpc.ExecutionResult, idx: int = 0) -> noderpc.StackItem:
    """
    Fetch the stack item at `idx` from the result stack. Performs basic validation and bounds checking.

    Args:
        res: execution result.
        idx: the index in the result stack to fetch the stack item from.
    """
    check_state_ok(res)
    if idx > len(res.stack) - 1:
        raise ValueError("Too few result items")
    return res.stack[idx]



"""
Containers following the NEP-6 wallet standard.
"""
from __future__ import annotations
import os.path
import json
from typing import Any, Optional
from collections.abc import Sequence
from jsonschema import validate  # type: ignore
from neo3.core import interfaces, cryptography
from neo3.contracts import contract, utils as contractutils
from neo3.wallet import account
from neo3.wallet import scrypt_parameters as scrypt


class Wallet(interfaces.IJson):
    """
    Base container.
    """

    _wallet_version = "1.0"

    # Wallet JSON validation schema
    json_schema = {
        "type": "object",
        "properties": {
            "name": {"type": ["string", "null"]},
            "version": {"type": "string"},
            "scrypt": scrypt.ScryptParameters.json_schema,
            "accounts": {
                "type": "array",
                "items": account.Account._json_schema,
                "minItems": 0,
            },
            "extra": {
                "type": ["object", "null"],
                "properties": {},
                "additionalProperties": True,
            },
        },
        "required": ["name", "version", "scrypt", "accounts", "extra"],
    }

    def __init__(
        self,
        name: Optional[str] = None,
        version: str = _wallet_version,
        scrypt_params: Optional[scrypt.ScryptParameters] = None,
        accounts: Optional[Sequence[account.Account]] = None,
        default_account: Optional[account.Account] = None,
        extra: Optional[dict[Any, Any]] = None,
    ):
        """
        Args:
            name: a user defined label for the wallet.
            version: the wallet's version, must be equal to or greater than 3.0.
            scrypt_params:  the parameters of the Scrypt algorithm used for encrypting and decrypting the private keys
            in the wallet.
            accounts: an array of Account objects to add to the wallet.
            extra: a user defined object for storing extra data. This field can be None.
        """

        self.name = name
        self.version = version
        try:
            self.scrypt = scrypt_params if scrypt_params else scrypt.ScryptParameters()
        except AttributeError:
            pass

        if accounts is None:
            accounts = []
        else:
            accounts = list(accounts)

        if default_account is not None and default_account not in accounts:
            # default account must be in the account list
            accounts.append(default_account)
        elif default_account is None and len(accounts) > 0:
            # if no account is defined as default, the first will be considered the default account
            default_account = accounts[0]

        self.accounts = accounts if accounts is not None else []
        self._default_account: Optional[account.Account] = default_account
        self.extra = extra if extra else {}

    def account_new(
        self, password: str, label: Optional[str] = None, is_default=False
    ) -> account.Account:
        """
        Create a new account and adds it in the wallet.

        Args:
            password: the password to encrypt the account.
            label: optional label to identify the account.
            is_default: set the created account as the default.
        """
        account_ = account.Account(
            password=password,
            watch_only=False,
            label=label,
            scrypt_parameters=self.scrypt,
        )

        self.account_add(account_, is_default)
        return account_

    @property
    def account_default(self) -> Optional[account.Account]:
        """
        Return the default account if at least one account is present.
        """
        return self._default_account

    def import_multisig_address(
        self, signing_threshold: int, public_keys: Sequence[cryptography.ECPoint]
    ) -> account.Account:
        """
        Import a multi-signature account into the container.

        Args:
            signing_threshold: minimum number of keys required for signing.
            public_keys: the public keys the multisignature address consists off.

        Raises:
            ValueError: if the signing treshold exceeds 1024.
                        if the signing treshold exceeds the number of public_keys.
        """
        if signing_threshold < 1 or signing_threshold > 1024:
            raise ValueError("Invalid signing threshold")

        if signing_threshold > len(public_keys):
            raise ValueError(
                f"Minimum signing threshold is {signing_threshold}, "
                f"received only {len(public_keys)} public keys"
            )

        multisig_contract = contract.Contract.create_multisig_contract(
            signing_threshold, public_keys
        )
        # we start with a watchonly as base
        account_ = account.Account.watch_only(multisig_contract.script_hash)
        account_.contract = account.AccountContract.from_contract(multisig_contract)

        # if the wallet contains an account matching one required in the multisig, then copy key material
        self._augment_multisig_with_key_material(account_)
        self.account_add(account_)
        return account_

    def _augment_multisig_with_key_material(self, acc: account.Account):
        """
        Tries to augment multi-sig accounts with key material such that they can be used for signing.

        There are 2 scenarios
        1. A multi-sig account is added while there already exists a regular account with key material for one of the
        keys required by the multi-signature.
        2. A regular account is added while there already exists a multi-sig account missing key material that the
        regular now adds.
        """
        if acc.contract is None:
            return

        is_multisig, _, public_keys = contractutils.parse_as_multisig_contract(
            acc.contract.script
        )
        if is_multisig:
            # scenario 1
            for account_ in self.accounts:
                if not account_.is_watchonly and account_.public_key in public_keys:
                    # copy key information of the first matching account
                    acc.encrypted_key = account_.encrypted_key
                    acc.public_key = account_.public_key
                    acc.scrypt_parameters = account_.scrypt_parameters
                    break
        else:
            # scenario 2
            for account_ in self.accounts:
                # testing acc.contract to silence mypy
                if account_.is_watchonly and account_.is_multisig and account_.contract:
                    _, _, public_keys = contractutils.parse_as_multisig_contract(
                        account_.contract.script
                    )
                    if acc.public_key in public_keys:
                        account_.encrypted_key = acc.encrypted_key
                        account_.public_key = acc.public_key
                        acc.scrypt_parameters = account_.scrypt_parameters
                        break

    def account_add(self, acc: account.Account, is_default=False) -> bool:
        """
        Add account in the wallet.

        Args:
            acc: the account to be added.
            is_default: set the created account as the default.

        Raises:
            ValueError: if the account's label is already used by another one.
        """
        # true if ok, false if duplicate (any other possible reasons? otherwise we need to throw exceptions)
        if acc in self.accounts:
            return False

        if acc.label is not None and self.account_get_by_label(acc.label) is not None:
            raise ValueError(f"Label is already used by an account '{acc.label}'")

        # if first account, also set to default
        if is_default or len(self.accounts) == 0:
            self._default_account = acc

        self._augment_multisig_with_key_material(acc)
        self.accounts.append(acc)
        return True

    def account_delete(self, acc: account.Account) -> bool:
        """
        Remove an account from the wallet.

        Args:
            acc: the account to be removed.
        """
        # return success or not
        if acc not in self.accounts:
            return False

        self.accounts.remove(acc)
        if acc == self._default_account:
            # if it was the default account, select a new one
            # first by default
            self._default_account = self.accounts[0] if len(self.accounts) > 0 else None

        return True

    def account_delete_by_label(self, label: str) -> bool:
        """
        Remove an account from the wallet given its label.

        Args:
            label: unique identifier of the account.
        """
        # return success or not
        acc = self.account_get_by_label(label)

        if acc is None:
            # account with that label was not found
            return False

        return self.account_delete(acc)

    def account_get_by_label(self, label: str) -> Optional[account.Account]:
        """
        Get an account given its label. Returns None if not found.

        Args:
            label: unique identifier of the account.
        """
        # returns the account with given label. None if the account is not found
        return next((acc for acc in self.accounts if acc.label == label), None)

    def save(self):
        """
        Save the wallet.

        This is called automatically when using the context manager.

        See Also:
            [DiskWallet](#neo3.wallet.wallet.DiskWallet)
        """
        pass

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        return {
            "name": self.name,
            "version": self.version,
            "scrypt": self.scrypt.to_json(),
            "accounts": [self._account_to_json(acc) for acc in self.accounts],
            "extra": self.extra if len(self.extra) > 0 else None,
        }

    def _account_to_json(self, acc: account.Account) -> dict:
        is_default = (
            self._default_account is not None
            and self._default_account.address == acc.address
        )
        json_account = acc.to_json()
        json_account["isDefault"] = is_default
        return json_account

    @classmethod
    def from_json(cls, json: dict, passwords: Optional[list[str]] = None):
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.
            passwords: the password to decrypt the account data.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the `version` property is under 1.0 or is not a valid string.
        """
        validate(json, schema=cls.json_schema)
        try:
            if float(json["version"]) < 1.0:
                raise ValueError("Format error - invalid 'version'")
        except ValueError:
            raise ValueError("Format error - invalid 'version'")

        accounts = []
        default_account = None
        scryptp = scrypt.ScryptParameters.from_json(json["scrypt"])
        if (len_accounts := len(json["accounts"])) > 0:
            if passwords is None:
                # mypy is being annoying about re-assigning to the same variable with a different type
                # we just need a list with None's for zip() to work and to create a watch only account.
                passwords = [None] * len(json["accounts"])  # type: ignore
            else:
                len_pws = len(passwords)
                if len_accounts != len_pws:
                    raise ValueError(
                        f"Incorrect number of passwords provided ({len_pws}) for number of accounts in wallet ({len_accounts})"
                    )
            for json_account, pw in zip(json["accounts"], passwords):
                account_from_json = account.Account.from_json(
                    json_account, pw, scrypt_parameters=scryptp
                )
                accounts.append(account_from_json)
                if (
                    default_account is None
                    and hasattr(json, "isDefault")
                    and json["isDefault"]
                ):
                    default_account = account_from_json

        return cls(
            name=json["name"],
            version=json["version"],
            scrypt_params=scryptp,
            accounts=accounts,
            default_account=default_account,
            extra=json["extra"],
        )

    @classmethod
    def from_file(cls, path: str, passwords: Optional[list[str]] = None):
        """
        Load wallet from file.

        Args:
            path: path as passed to `open()`.
            passwords: the password to decrypt the account data.
        """
        with open(path, "r") as f:
            return cls.from_json(json.load(f), passwords)

    def __enter__(self) -> Wallet:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.save()


class DiskWallet(Wallet):
    """
    Specialised wallet for persisting to media.
    """

    _default_path = "./wallet.json"

    def __init__(
        self,
        path: str,
        name: Optional[str] = None,
        version: str = Wallet._wallet_version,
        scrypt_params: Optional[scrypt.ScryptParameters] = None,
        accounts: Optional[Sequence[account.Account]] = None,
        default_account: Optional[account.Account] = None,
        extra: Optional[dict] = None,
    ):
        """
        Args:
            path: the location where the wallet will be stored.
            name: a user defined label for the wallet.
            version: the wallet's version, must be equal to or greater than 3.0.
            scrypt_params:  the parameters of the Scrypt algorithm used for encrypting and decrypting the private keys
            in the wallet.
            accounts: an array of Account objects to add to the wallet.
            extra: a user defined object for storing extra data. This field can be None.
        """

        filepath, extension = os.path.splitext(path)
        if len(extension) == 0:
            # if the path doesn't have a file extension, sets it as a .json file
            path += ".json"

        if name is None:
            # sets the wallet name the same as the file name
            dir_path, name = os.path.split(path)
            name, extension = os.path.splitext(name)

        self.path: str = path
        super(DiskWallet, self).__init__(
            name=name,
            version=version,
            scrypt_params=scrypt_params,
            accounts=accounts,
            default_account=default_account,
            extra=extra,
        )

    def save(self) -> None:
        """
        Persist the wallet to disk.
        """
        with open(self.path, "w") as json_file:
            json.dump(self.to_json(), json_file)

    @classmethod
    def default(
        cls, path: str = _default_path, name: Optional[str] = "wallet.json"
    ) -> DiskWallet:
        """
        Create a new wallet with the default settings.

        Args:
            path: the JSON's path.
            name: the wallet name.
        """
        return cls(
            path=path,
            name=name,
            version=cls._wallet_version,
            scrypt_params=scrypt.ScryptParameters(),
            accounts=[],
            extra=None,
        )

    @classmethod
    def from_json(cls, json: dict, passwords: Optional[list[str]] = None):
        w = Wallet.from_json(json, passwords)
        path = ""
        return cls(
            path, w.name, w.version, w.scrypt, w.accounts, w.account_default, w.extra
        )

    @classmethod
    def from_file(cls, path: str, passwords: Optional[list[str]] = None):
        with open(path, "r") as f:
            w = cls.from_json(json.load(f), passwords)
            w.path = path
            return w



"""
NEO Virtual Machine classes.
"""
from __future__ import annotations
import hashlib
from enum import IntEnum
from neo3.contracts import callflags
from neo3.core import types, serialization, cryptography
from typing import Optional, Iterator, Union, Type, Protocol
from collections.abc import Sequence


class StackItemType(IntEnum):
    """
    StackItemType as defined inside the virtual machine
    """

    ANY = 0x0
    POINTER = 0x10
    BOOLEAN = 0x20
    INTEGER = 0x21
    BYTESTRING = 0x28
    BUFFER = 0x30
    ARRAY = 0x40
    STRUCT = 0x41
    MAP = 0x48


def _syscall_name_to_int(name: str) -> int:
    return int.from_bytes(
        hashlib.sha256(name.encode()).digest()[:4], "little", signed=False
    )


class OpCode(IntEnum):
    """
    NEO Virtual Machine instructions.

    Can be concatenated into a hex-escaped bytes sequence.

    Example:
        In [1]: from neo3.contracts import vm

        In [2]: script = vm.OpCode.PUSHDATA1 + b'\x01' + vm.OpCode.RET

        In [3]: script
        Out[3]: b'\x0c\x01@'
    """

    PUSHINT8 = 0x00
    PUSHINT16 = 0x01
    PUSHINT32 = 0x02
    PUSHINT64 = 0x03
    PUSHINT128 = 0x04
    PUSHINT256 = 0x05
    PUSHT = 0x08
    PUSHF = 0x09
    PUSHA = 0x0A
    PUSHNULL = 0x0B
    PUSHDATA1 = 0x0C
    PUSHDATA2 = 0x0D
    PUSHDATA4 = 0x0E
    PUSHM1 = 0x0F
    PUSH0 = 0x10
    PUSH1 = 0x11
    PUSH2 = 0x12
    PUSH3 = 0x13
    PUSH4 = 0x14
    PUSH5 = 0x15
    PUSH6 = 0x16
    PUSH7 = 0x17
    PUSH8 = 0x18
    PUSH9 = 0x19
    PUSH10 = 0x1A
    PUSH11 = 0x1B
    PUSH12 = 0x1C
    PUSH13 = 0x1D
    PUSH14 = 0x1E
    PUSH15 = 0x1F
    PUSH16 = 0x20
    NOP = 0x21
    JMP = 0x22
    JMP_L = 0x23
    JMPIF = 0x24
    JMPIF_L = 0x25
    JMPIFNOT = 0x26
    JMPIFNOT_L = 0x27
    JMPEQ = 0x28
    JMPEQ_L = 0x29
    JMPNE = 0x2A
    JMPNE_L = 0x2B
    JMPGT = 0x2C
    JMPGT_L = 0x2D
    JMPGE = 0x2E
    JMPGE_L = 0x2F
    JMPLT = 0x30
    JMPLT_L = 0x31
    JMPLE = 0x32
    JMPLE_L = 0x33
    CALL = 0x34
    CALL_L = 0x35
    CALLA = 0x36
    CALLT = 0x37
    ABORT = 0x38
    ASSERT = 0x39
    THROW = 0x3A
    TRY = 0x3B
    TRY_L = 0x3C
    ENDTRY = 0x3D
    ENDTRY_L = 0x3E
    ENDFINALLY = 0x3F
    RET = 0x40
    SYSCALL = 0x41
    DEPTH = 0x43
    DROP = 0x45
    NIP = 0x46
    XDROP = 0x48
    CLEAR = 0x49
    DUP = 0x4A
    OVER = 0x4B
    PICK = 0x4D
    TUCK = 0x4E
    SWAP = 0x50
    ROT = 0x51
    ROLL = 0x52
    REVERSE3 = 0x53
    REVERSE4 = 0x54
    REVERSEN = 0x55
    INITSSLOT = 0x56
    INITSLOT = 0x57
    LDSFLD0 = 0x58
    LDSFLD1 = 0x59
    LDSFLD2 = 0x5A
    LDSFLD3 = 0x5B
    LDSFLD4 = 0x5C
    LDSFLD5 = 0x5D
    LDSFLD6 = 0x5E
    LDSFLD = 0x5F
    STSFLD0 = 0x60
    STSFLD1 = 0x61
    STSFLD2 = 0x62
    STSFLD3 = 0x63
    STSFLD4 = 0x64
    STSFLD5 = 0x65
    STSFLD6 = 0x66
    STSFLD = 0x67
    LDLOC0 = 0x68
    LDLOC1 = 0x69
    LDLOC2 = 0x6A
    LDLOC3 = 0x6B
    LDLOC4 = 0x6C
    LDLOC5 = 0x6D
    LDLOC6 = 0x6E
    LDLOC = 0x6F
    STLOC0 = 0x70
    STLOC1 = 0x71
    STLOC2 = 0x72
    STLOC3 = 0x73
    STLOC4 = 0x74
    STLOC5 = 0x75
    STLOC6 = 0x76
    STLOC = 0x77
    LDARG0 = 0x78
    LDARG1 = 0x79
    LDARG2 = 0x7A
    LDARG3 = 0x7B
    LDARG4 = 0x7C
    LDARG5 = 0x7D
    LDARG6 = 0x7E
    LDARG = 0x7F
    STARG0 = 0x80
    STARG1 = 0x81
    STARG2 = 0x82
    STARG3 = 0x83
    STARG4 = 0x84
    STARG5 = 0x85
    STARG6 = 0x86
    STARG = 0x87
    NEWBUFFER = 0x88
    MEMCPY = 0x89
    CAT = 0x8B
    SUBSTR = 0x8C
    LEFT = 0x8D
    RIGHT = 0x8E
    INVERT = 0x90
    AND = 0x91
    OR = 0x92
    XOR = 0x93
    EQUAL = 0x97
    NOTEQUAL = 0x98
    SIGN = 0x99
    ABS = 0x9A
    NEGATE = 0x9B
    INC = 0x9C
    DEC = 0x9D
    ADD = 0x9E
    SUB = 0x9F
    MUL = 0xA0
    DIV = 0xA1
    MOD = 0xA2
    POW = 0xA3
    SQRT = 0xA4
    MODMUL = 0xA5
    MODPOW = 0xA6
    SHL = 0xA8
    SHR = 0xA9
    NOT = 0xAA
    BOOLAND = 0xAB
    BOOLOR = 0xAC
    NZ = 0xB1
    NUMEQUAL = 0xB3
    NUMNOTEQUAL = 0xB4
    LT = 0xB5
    LE = 0xB6
    GT = 0xB7
    GE = 0xB8
    MIN = 0xB9
    MAX = 0xBA
    WITHIN = 0xBB
    PACKMAP = 0xBE
    PACKSTRUCT = 0xBF
    PACK = 0xC0
    UNPACK = 0xC1
    NEWARRAY0 = 0xC2
    NEWARRAY = 0xC3
    NEWARRAY_T = 0xC4
    NEWSTRUCT0 = 0xC5
    NEWSTRUCT = 0xC6
    NEWMAP = 0xC8
    SIZE = 0xCA
    HASKEY = 0xCB
    KEYS = 0xCC
    VALUES = 0xCD
    PICKITEM = 0xCE
    APPEND = 0xCF
    SETITEM = 0xD0
    REVERSEITEMS = 0xD1
    REMOVE = 0xD2
    CLEARITEMS = 0xD3
    POPITEM = 0xD4
    ISNULL = 0xD8
    ISTYPE = 0xD9
    CONVERT = 0xDB

    def __eq__(self, other):
        if super(OpCode, self).__eq__(other) is True:
            return True
        if isinstance(other, bytes):
            return self.value.to_bytes(1, "little") == other
        return False

    def __add__(self, other):
        if isinstance(other, bytes):
            return self.value.to_bytes(1, "little") + other
        elif isinstance(other, OpCode):
            return self.value.to_bytes(1, "little") + other.to_bytes(1, "little")
        else:
            return super(OpCode, self).__add__(other)

    def __radd__(self, other):
        if isinstance(other, bytes):
            return other + self.value.to_bytes(1, "little")
        elif isinstance(other, OpCode):
            return other.to_bytes(1, "little") + self.value.to_bytes(1, "little")
        else:
            return super(OpCode, self).__radd__(other)


ContractParameter = Union[
    bool,
    int,
    str,
    bytes,
    bytearray,
    types.BigInteger,
    types.UInt160,
    types.UInt256,
    cryptography.ECPoint,
    "ContractParameterArray",
    "ContractParameterDict",
    Type[serialization.ISerializable_T],
]


class ContractParameterArray(Protocol):
    """"""

    def insert(self, index: int, value: ContractParameter) -> None:
        ...

    def __getitem__(self, i: int) -> ContractParameter:
        ...

    def __setitem__(self, i: int, o: ContractParameter) -> None:
        ...

    def __delitem__(self, i: int) -> None:
        ...


class ContractParameterDict(Protocol):
    """"""

    def __setitem__(self, k: ContractParameter, v: ContractParameter) -> None:
        ...

    def __delitem__(self, v: ContractParameter) -> None:
        ...

    def __getitem__(self, k: ContractParameter) -> ContractParameter:
        ...

    def __iter__(self) -> Iterator[ContractParameter]:
        ...


class ScriptBuilder:
    """
    A utility class to create scripts (sequence of opcodes) that can be executed by the
    NEO Virtual Machine.
    """

    def __init__(self):
        self.data = bytearray()

    def emit(self, opcode: OpCode, data: Optional[bytes] = None) -> ScriptBuilder:
        self.emit_raw(opcode.value.to_bytes(1, "little"))
        if data is not None:
            self.emit_raw(data)
        return self

    def emit_push(self, value) -> ScriptBuilder:
        if value is None:
            return self.emit(OpCode.PUSHNULL)
        elif isinstance(value, bool):
            if value is True:
                return self.emit(OpCode.PUSHT)
            else:
                return self.emit(OpCode.PUSHF)
        elif isinstance(value, str):
            self.emit_push(value.encode("utf-8"))
            return self
        elif isinstance(value, serialization.ISerializable):
            return self.emit_push(value.to_array())
        elif isinstance(value, IntEnum):
            return self.emit_push(value.value)
        elif isinstance(value, (types.BigInteger, int)):
            if -1 <= value <= 16:
                self.emit_raw((OpCode.PUSH0 + value).to_bytes(1, "little"))
                return self
            else:
                if isinstance(value, int):
                    bigint = types.BigInteger(value)
                else:
                    bigint = value
                data = bytearray(bigint.to_array())
                if len(data) == 1:
                    return self.emit(OpCode.PUSHINT8, data)
                if len(data) == 2:
                    return self.emit(OpCode.PUSHINT16, data)
                if len(data) <= 4:
                    self._pad_right(data, 4, bigint.sign < 0)
                    return self.emit(OpCode.PUSHINT32, data)
                if len(data) <= 8:
                    self._pad_right(data, 8, bigint.sign < 0)
                    return self.emit(OpCode.PUSHINT64, data)
                if len(data) <= 16:
                    self._pad_right(data, 16, bigint.sign < 0)
                    return self.emit(OpCode.PUSHINT128, data)
                if len(data) <= 32:
                    self._pad_right(data, 32, bigint.sign < 0)
                    return self.emit(OpCode.PUSHINT256, data)
                raise ValueError("Input number exceeds maximum data size of 32 bytes")
        elif isinstance(value, (bytes, bytearray)):
            len_value = len(value)
            if len_value > 0xFFFFFFFF:
                raise ValueError(
                    f"Value is too long {len_value}. Maximum allowed length is 0xFFFF_FFFF"
                )

            if len_value < 0x100:
                self.emit(OpCode.PUSHDATA1)
                self.emit_raw(len_value.to_bytes(1, "little"))
                self.emit_raw(value)
            elif len_value < 0x10000:
                self.emit(OpCode.PUSHDATA2)
                self.emit_raw((len_value & 0xFF).to_bytes(1, "little"))
                self.emit_raw(((len_value >> 8) & 0xFF).to_bytes(1, "little"))
                self.emit_raw(value)
            else:
                self.emit(OpCode.PUSHDATA4)
                self.emit_raw(len_value.to_bytes(4, "little"))
                self.emit_raw(value)
            return self
        elif isinstance(value, Sequence):
            for item in reversed(value):
                if isinstance(item, Sequence):
                    self.emit_push(item)
                    continue
                self.emit_push(item)
            self.emit_push(len(value))
            self.emit(OpCode.PACK)
            return self
        elif isinstance(value, dict):
            for k, v in reversed(value.items()):
                # This restriction exists on the VM side where keys to a 'Map' may only be of 'PrimitiveType'
                if not isinstance(
                    k, (int, str, bool, bytes, serialization.ISerializable)
                ):
                    raise ValueError(
                        f"Unsupported key type {type(k)}. "
                        f"Supported types by the VM are bool, int, str, bytes or ISerializable"
                    )
                self.emit_push(v)
                self.emit_push(k)
            self.emit_push(len(value))
            return self.emit(OpCode.PACKMAP)
        else:
            raise ValueError(f"Unsupported value type {type(value)}")

    def emit_raw(self, data: bytes) -> ScriptBuilder:
        self.data.extend(data)
        return self

    def emit_jump(self, opcode: OpCode, offset: int) -> ScriptBuilder:
        if opcode < OpCode.JMP or opcode > OpCode.JMPLE_L:
            raise ValueError(f"OpCode {opcode.name} is not a valid jump OpCode")

        # auto correct opcode
        if opcode % 2 == 0 and (offset < -128 or offset > 127):
            opcode = OpCode(opcode + 1)

        if opcode % 2 == 0:
            return self.emit(opcode, offset.to_bytes(1, "little", signed=True))
        else:
            return self.emit(opcode, offset.to_bytes(4, "little", signed=True))

    def emit_call(self, offset: int) -> ScriptBuilder:
        if offset < -128 or offset > 127:
            return self.emit(OpCode.CALL_L, offset.to_bytes(4, "little"))
        else:
            return self.emit(OpCode.CALL, offset.to_bytes(1, "little"))

    def emit_syscall(self, syscall: int | Syscall):
        if isinstance(syscall, Syscall):
            syscall = syscall.number

        return self.emit(OpCode.SYSCALL, syscall.to_bytes(4, "little"))

    def emit_contract_call(
        self,
        script_hash: types.UInt160,
        operation: str,
        call_flags: Optional[callflags.CallFlags] = None,
    ) -> ScriptBuilder:
        """
        Emit opcode sequence to call a smart contrat operation.

        Args:
            script_hash: contract script hash.
            operation: method to call on contract.
            call_flags: call flags for the operation.
        """
        self.emit(OpCode.NEWARRAY0)
        self.emit_push(callflags.CallFlags.ALL if call_flags is None else call_flags)
        self.emit_push(operation)
        self.emit_push(script_hash)
        self.emit_syscall(Syscalls.SYSTEM_CONTRACT_CALL)
        return self

    def emit_contract_call_with_args(
        self,
        script_hash: types.UInt160,
        operation: str,
        args: ContractParameter,
        call_flags: Optional[callflags.CallFlags] = None,
    ) -> ScriptBuilder:
        """
        Emit opcode sequence to call a smart contrat operation with arguments.

        Args:
            script_hash: contract script hash.
            operation: method to call on contract.
            args: parameters to pass to the `operation`.
            call_flags: call flags for the operation.
        """
        if isinstance(args, Sequence):
            for arg in reversed(args):
                self.emit_push(arg)
            self.emit_push(len(args))
            self.emit(OpCode.PACK)
        else:
            self.emit_push(args)
        self.emit_push(callflags.CallFlags.ALL if call_flags is None else call_flags)
        self.emit_push(operation)
        self.emit_push(script_hash)
        self.emit_syscall(Syscalls.SYSTEM_CONTRACT_CALL)
        return self

    def emit_contract_call_and_count_iterator(
        self,
        script_hash,
        operation: str,
        call_flags: Optional[callflags.CallFlags] = None,
    ) -> ScriptBuilder:
        return self._emit_contract_call_and_count_iterator(
            script_hash, operation, None, call_flags
        )

    def emit_contract_call_and_unwrap_iterator(
        self,
        script_hash,
        operation: str,
        call_flags: Optional[callflags.CallFlags] = None,
        unwrap_limit: int = 2000,
        start_index: int = 0,
    ) -> ScriptBuilder:
        return self._emit_contract_call_and_unwrap_iterator(
            script_hash, operation, None, call_flags, unwrap_limit, start_index
        )

    def emit_contract_call_with_args_and_count_iterator(
        self,
        script_hash,
        operation: str,
        args: Optional[ContractParameter] = None,
        call_flags: Optional[callflags.CallFlags] = None,
    ) -> ScriptBuilder:
        return self._emit_contract_call_and_count_iterator(
            script_hash, operation, args, call_flags
        )

    def emit_contract_call_with_args_and_unwrap_iterator(
        self,
        script_hash,
        operation: str,
        args: ContractParameter,
        call_flags: Optional[callflags.CallFlags] = None,
    ) -> ScriptBuilder:
        return self._emit_contract_call_and_unwrap_iterator(
            script_hash, operation, args, call_flags
        )

    def to_array(self) -> bytes:
        return bytes(self.data)

    def _emit_contract_call_and_unwrap_iterator(
        self,
        script_hash,
        operation: str,
        args: Optional[ContractParameter] = None,
        call_flags: Optional[callflags.CallFlags] = None,
        unwrap_limit: int = 2000,
        start_index: int = 0,
    ) -> ScriptBuilder:
        """

        Args:
            script_hash: contract hash to call
            operation: method name to call
            args: arguments passed to the method called
            call_flags: call flags of the method called
            unwrap_limit: maximum number of items to return. Can't be larger than the MaxStackSize limit configured for
            the VM or it will throw an exception.
            https://github.com/neo-project/neo-vm/blob/5b0a39811b34abacab1273f3ee5a9a9f7e52ac7b/src/Neo.VM/ExecutionEngineLimits.cs#L34C21-L34C33
            The current default is slightly lower than the max, because that allows for a few other items to be on the
            stack in other function frames.
            start_index: index in the iterator to start capturing values from
        """
        # jump to local variables initialization code
        self.emit_jump(OpCode.JMP, 4)
        # the following 2 instructions are for loading the result array and exiting the script
        # it is at the beginning because it make it easy to calculate the offset
        return_results = len(self.data)
        self.emit(OpCode.LDLOC0)
        self.emit(OpCode.RET)
        # reserve local variables for the `iterator`, the `results` list, the `start_index` and a `counter`
        self.emit(OpCode.INITSLOT)
        self.emit_raw(b"\x04")
        self.emit_raw(b"\x00")
        # store results list in pos 0
        self.emit(OpCode.NEWARRAY0)
        self.emit(OpCode.STLOC0)

        if args is None or (isinstance(args, Sequence) and len(args) == 0):
            self.emit_contract_call(script_hash, operation, call_flags)
        else:
            self.emit_contract_call_with_args(script_hash, operation, args, call_flags)
        # store iterator in pos 1
        self.emit(OpCode.STLOC1)
        # store stack item counter in pos 2
        self.emit_push(0)
        self.emit(OpCode.STLOC2)
        # store the start index in pos 3
        self.emit_push(start_index)
        self.emit(OpCode.STLOC3)
        """
        Next set of opcodes does the following

        while iterator.next()
          if ctr >= start_index:
            results.append(iterator.value)
          ctr += 1
          if ctr == end_index:
            break
        return results
        """
        loop_start = len(self.data)
        # load iterator as argument for iterator.next
        self.emit(OpCode.LDLOC1)
        # test if the iterator has a value
        self.emit_syscall(Syscalls.SYSTEM_ITERATOR_NEXT)
        # if not jump to exit routine
        self.emit_jump(OpCode.JMPIFNOT, self._offset_to(return_results))
        # load item counter + start_index
        self.emit(OpCode.LDLOC2)
        self.emit(OpCode.LDLOC3)
        self.emit(OpCode.GE)
        skip_append_value_to_array = (
            len(self.data) + 11
        )  # 11 is all the instructions + operands until the 'increase counter' label
        self.emit_jump(OpCode.JMPIFNOT, self._offset_to(skip_append_value_to_array))
        # load iterator as argument for iterator.value
        self.emit(OpCode.LDLOC1)
        # get result
        self.emit_syscall(Syscalls.SYSTEM_ITERATOR_VALUE)
        # load array
        self.emit(OpCode.LDLOC0)
        # fix argument order for APPEND
        self.emit(OpCode.SWAP)
        self.emit(OpCode.APPEND)
        # increase counter
        # load stack item counter
        self.emit(OpCode.LDLOC2)
        self.emit(OpCode.INC)
        self.emit(OpCode.DUP)
        self.emit(OpCode.STLOC2)
        # load end_index
        self.emit_push(start_index + unwrap_limit)
        self.emit(OpCode.NUMEQUAL)
        self.emit_jump(OpCode.JMPIF, self._offset_to(return_results))
        # jump back to start of `while` loop
        self.emit_jump(OpCode.JMP, self._offset_to(loop_start))
        return self

    def _emit_contract_call_and_count_iterator(
        self,
        script_hash,
        operation: str,
        args: Optional[ContractParameter] = None,
        call_flags: Optional[callflags.CallFlags] = None,
    ) -> ScriptBuilder:
        """
        Count the number of values returned by the iterator

        Args:
            script_hash: contract hash to call
            operation: method name to call
            args: arguments passed to the method called
            call_flags: call flags of the method called
        """
        # jump to local variables initialization code
        self.emit_jump(OpCode.JMP, 4)
        # the following 2 instructions are for loading the counter and exiting the script
        # it is at the beginning because it makes it easy to calculate the offset
        return_results = len(self.data)
        self.emit(OpCode.LDLOC1)
        self.emit(OpCode.RET)
        # reserve local variables for the `iterator` and a `counter`
        self.emit(OpCode.INITSLOT)
        self.emit_raw(b"\x02")
        self.emit_raw(b"\x00")

        if args is None or (isinstance(args, Sequence) and len(args) == 0):
            self.emit_contract_call(script_hash, operation, call_flags)
        else:
            self.emit_contract_call_with_args(script_hash, operation, args, call_flags)
        # store iterator in pos 0
        self.emit(OpCode.STLOC0)

        # store stack item counter in pos 1
        self.emit_push(0)
        self.emit(OpCode.STLOC1)
        """
        Next set of opcodes does the following

        while iterator.next()
          ctr += 1
        return results
        """
        loop_start = len(self.data)
        # load iterator as argument for iterator.next
        self.emit(OpCode.LDLOC0)
        # test if the iterator has a value
        self.emit_syscall(Syscalls.SYSTEM_ITERATOR_NEXT)
        # if not jump to exit routine
        self.emit_jump(OpCode.JMPIFNOT, self._offset_to(return_results))
        self.emit_push("it next")
        self.emit_syscall(Syscalls.SYSTEM_RUNTIME_LOG)
        # load stack item counter
        self.emit(OpCode.LDLOC1)
        self.emit(OpCode.INC)
        self.emit(OpCode.STLOC1)
        # jump back to start of `while` loop
        self.emit_jump(OpCode.JMP, self._offset_to(loop_start))
        return self

    def _pad_right(self, data: bytearray, length: int, is_negative: bool):
        if len(data) >= length:
            return
        pad = b"\xFF" if is_negative else b"\x00"
        while len(data) != length:
            data.extend(pad)

    def _offset_to(self, absolute_position: int):
        return absolute_position - len(self.data)


class VMState(IntEnum):
    NONE = 0
    HALT = 1 << 0
    FAULT = 1 << 1
    BREAK = 1 << 2

    @staticmethod
    def from_string(value: str) -> VMState:
        match value:
            case "NONE":
                return VMState.NONE
            case "HALT":
                return VMState.HALT
            case "FAULT":
                return VMState.FAULT
            case "BREAK":
                return VMState.BREAK
            case _:
                raise ValueError(f"{value} cannot be converted to VMState")


class Syscall:
    def __init__(self, syscall_name: str, required_callflags: callflags.CallFlags):
        self.name = syscall_name
        self.number = _syscall_name_to_int(self.name)
        self.required_callflags = required_callflags

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.number}, {self.name}>"

    def __eq__(self, other):
        if type(other) == int:
            return self.number == other
        if type(other) == str:
            return self.name == other
        if type(other) == type(self):
            return self.name == other.name and self.number == other.number
        if type(other) in (bytes, bytearray):
            return self.to_array() == other
        else:
            return False

    def to_array(self) -> bytes:
        return self.number.to_bytes(4, "little")


class Syscalls:
    """
    Container holding all NEO blockchain interop syscalls.
    """

    #: Call another smart contract.
    SYSTEM_CONTRACT_CALL = Syscall(
        "System.Contract.Call",
        callflags.CallFlags.READ_STATES | callflags.CallFlags.ALLOW_CALL,
    )
    #: Internal use only. Added for completeness.
    SYSTEM_CONTRACT_CALL_NATIVE = Syscall(
        "System.Contract.CallNative", callflags.CallFlags.NONE
    )
    #: Get the call flags for the current execution context text.
    SYSTEM_CONTRACT_GET_CALL_FLAGS = Syscall(
        "System.Contract.GetCallFlags", callflags.CallFlags.NONE
    )
    #: Get the (contract) account scripthash for the given public key.
    SYSTEM_CONTRACT_CREATE_STANDARD_ACCOUNT = Syscall(
        "System.Contract.CreateStandardAccount", callflags.CallFlags.NONE
    )
    #: Get the (multisignature contract) account scripthash for the given public key(s).
    SYSTEM_CONTRACT_CREATE_MULTI_SIGNATURE_ACCOUNT = Syscall(
        "System.Contract.CreateMultisigAccount", callflags.CallFlags.NONE
    )
    #: Internal use only. Added for completeness.
    SYSTEM_CONTRACT_NATIVE_ON_PERSIST = Syscall(
        "System.Contract.NativeOnPersist", callflags.CallFlags.STATES
    )
    #: Internal use only. Added for completeness.
    SYSTEM_CONTRACT_NATIVE_POST_PERSIST = Syscall(
        "System.Contract.NativePostPersist", callflags.CallFlags.STATES
    )

    #: Validates the signature of the current script container (usually a transaction).
    SYSTEM_CRYPTO_CHECK_STANDARD_ACCOUNT = Syscall(
        "System.Crypto.CheckSig", callflags.CallFlags.NONE
    )
    #: Validates the signatures of the current script container (usually a transaction).
    SYSTEM_CRYPTO_CHECK_MULTI_SIGNATURE_ACCOUNT = Syscall(
        "System.Crypto.CheckMultisig", callflags.CallFlags.NONE
    )

    #: Advance the iterator to the next element of the collection. See also SYSTEM_STORAGE_FIND.
    SYSTEM_ITERATOR_NEXT = Syscall("System.Iterator.Next", callflags.CallFlags.NONE)
    #: Get the element in the collection at the current position of the iterator. See also SYSTEM_STORAGE_FIND.
    SYSTEM_ITERATOR_VALUE = Syscall("System.Iterator.Value", callflags.CallFlags.NONE)

    #: Get the name of the current platform. For NEO blockchain fixed to "NEO".
    SYSTEM_RUNTIME_PLATFORM = Syscall(
        "System.Runtime.Platform", callflags.CallFlags.NONE
    )
    #: Get the Signers of the current transaction.
    SYSTEM_RUNTIME_CURRENT_SIGNERS = Syscall(
        "System.Runtime.CurrentSigners", callflags.CallFlags.NONE
    )
    #: Get the network magic number.
    SYSTEM_RUNTIME_GET_NETWORK = Syscall(
        "System.Runtime.GetNetwork", callflags.CallFlags.NONE
    )
    #: Get the address version.
    SYSTEM_RUNTIME_GET_ADDRESS_VERSION = Syscall(
        "System.Runtime.GetAddressVersion", callflags.CallFlags.NONE
    )
    #: Get the trigger type used in the engine for the current execution.
    SYSTEM_RUNTIME_GET_TRIGGER = Syscall(
        "System.Runtime.GetTrigger", callflags.CallFlags.NONE
    )
    #: Get the timestamp of the block currently being persisted.
    SYSTEM_RUNTIME_GET_TIME = Syscall(
        "System.Runtime.GetTime", callflags.CallFlags.NONE
    )
    #: Get the script container of the current execution (usually the transaction).
    SYSTEM_RUNTIME_GET_SCRIPT_CONTAINER = Syscall(
        "System.Runtime.GetScriptContainer", callflags.CallFlags.NONE
    )
    #: Get the script hash of the current execution context.
    SYSTEM_RUNTIME_GET_EXECUTING_SCRIPT_HASH = Syscall(
        "System.Runtime.GetExecutingScriptHash", callflags.CallFlags.NONE
    )
    #: Get the script hash of the calling contract.
    SYSTEM_RUNTIME_GET_CALLING_SCRIPT_HASH = Syscall(
        "System.Runtime.GetCallingScriptHash", callflags.CallFlags.NONE
    )
    #: Get the script hash of the first execution context script. For a transaction this equals to `Transaction.script`.
    SYSTEM_RUNTIME_GET_ENTRY_SCRIPT_HASH = Syscall(
        "System.Runtime.GetEntryScriptHash", callflags.CallFlags.NONE
    )
    #: Validate whether the specified account has witnessed the current transaction.
    SYSTEM_RUNTIME_CHECK_WITNESS = Syscall(
        "System.Runtime.CheckWitness", callflags.CallFlags.NONE
    )
    #: Get the number of times the current contract has been called during the execution.
    SYSTEM_RUNTIME_GET_INVOCATION_COUNTER = Syscall(
        "System.Runtime.GetInvocationCounter", callflags.CallFlags.NONE
    )
    #: Get a random number.
    SYSTEM_RUNTIME_GET_RANDOM = Syscall(
        "System.Runtime.GetRandom", callflags.CallFlags.NONE
    )
    #: Write a log message.
    SYSTEM_RUNTIME_LOG = Syscall("System.Runtime.Log", callflags.CallFlags.ALLOW_NOTIFY)
    #: Send a notification.
    SYSTEM_RUNTIME_NOTIFY = Syscall(
        "System.Runtime.Notify", callflags.CallFlags.ALLOW_NOTIFY
    )
    #: Get the list of notifications sent by the specified contract during the execution.
    SYSTEM_RUNTIME_GET_NOTIFICATIONS = Syscall(
        "System.Runtime.GetNotifications", callflags.CallFlags.NONE
    )
    #: Get the remaining GAS that can be spent in order to complete the execution.
    SYSTEM_RUNTIME_GAS_LEFT = Syscall(
        "System.Runtime.GasLeft", callflags.CallFlags.NONE
    )
    #: Burns gas.
    SYSTEM_RUNTIME_BURN_GAS = Syscall(
        "System.Runtime.BurnGas", callflags.CallFlags.NONE
    )

    #: Get the storage context for the current contract.
    SYSTEM_STORAGE_GET_CONTEXT = Syscall(
        "System.Storage.GetContext", callflags.CallFlags.READ_STATES
    )
    #: Get the storage context for the current contract as read-only.
    SYSTEM_STORAGE_GET_READ_ONLY_CONTEXT = Syscall(
        "System.Storage.GetReadOnlyContext", callflags.CallFlags.READ_STATES
    )
    #: Convert the existing context to a new read-only context.
    SYSTEM_STORAGE_AS_READ_ONLY = Syscall(
        "System.Storage.AsReadOnly", callflags.CallFlags.READ_STATES
    )
    #: Get an entry from storage by a specified key.
    SYSTEM_STORAGE_GET = Syscall("System.Storage.Get", callflags.CallFlags.READ_STATES)
    #: Find entries from storage by a given a search prefix and search options.
    SYSTEM_STORAGE_FIND = Syscall(
        "System.Storage.Find", callflags.CallFlags.READ_STATES
    )
    #: Persist an entry to storage under a specified key.
    SYSTEM_STORAGE_PUT = Syscall("System.Storage.Put", callflags.CallFlags.WRITE_STATES)
    #: Delete an entry from storage under a specified key.
    SYSTEM_STORAGE_DELETE = Syscall(
        "System.Storage.Delete", callflags.CallFlags.WRITE_STATES
    )

    @classmethod
    def all(cls) -> Iterator[Syscall]:
        for name, value in vars(cls).items():
            if name.isupper():
                yield value

    @classmethod
    def get_by_number(cls, syscall_number: int) -> Optional[Syscall]:
        for name, value in vars(cls).items():
            if name.isupper() and value.number == syscall_number:
                return value
        else:
            return None

    @classmethod
    def get_by_name(cls, syscall_name: str) -> Optional[Syscall]:
        for name, value in vars(cls).items():
            if name.isupper() and value.name == syscall_name:
                return value
        else:
            return None



"""
Signing functions for use with `ChainFacade.invoke`.
"""
import os
from dataclasses import dataclass
from neo3.network.payloads import transaction
from typing import Callable, Awaitable
from neo3.wallet import account


@dataclass
class SigningDetails:
    network: int
    #: for which Signer we're adding a witness
    # witness_index: int


# The signing function adds a witness to the provided transaction
# SigningFunction = Callable[[transaction.Transaction, SigningDetails], None]
SigningFunction = Callable[[transaction.Transaction, SigningDetails], Awaitable]


def sign_insecure_with_account(acc: account.Account, password: str) -> SigningFunction:
    """
    Sign and add a witness using the account and the provided account password.
    """

    async def insecure_account_signer(
        tx: transaction.Transaction, details: SigningDetails
    ):
        # this will automatically add a witness
        acc.sign_tx(tx, password, details.network)

    return insecure_account_signer


def sign_secure_with_account(
    acc: account.Account, env_password_name: str
) -> SigningFunction:
    """
    Sign and add a witness using the account. The account password is read from the environment variables.
    """

    async def insecure_account_signer(
        tx: transaction.Transaction, details: SigningDetails
    ):
        # this will automatically add a witness
        acc.sign_tx(tx, os.environ[env_password_name], details.network)

    return insecure_account_signer


def sign_secure_with_ledger() -> SigningFunction:
    raise NotImplementedError


def sign_on_remote_server() -> SigningFunction:
    async def remote_server_signer(
        tx: transaction.Transaction, details: SigningDetails
    ):
        # call some remote API to get the signature
        # and add a witness with the signature
        raise NotImplementedError

    return remote_server_signer


def sign_insecure_with_multisig_account(
    acc: account.Account, password: str
) -> SigningFunction:
    """
    Sign and add a multi-signature witness.

    This only works for a 1 out of n multi-signature account.

    Args:
        acc: a multi-signature account
        password: the password of the account to sign with
    """

    async def insecure_account_signer(
        tx: transaction.Transaction, details: SigningDetails
    ):
        ctx = account.MultiSigContext()
        # this will automatically add a witness
        acc.sign_multisig_tx(tx, password, ctx, details.network)

    return insecure_account_signer



import unittest
from neo3.contracts import abi


class ContractParameterDefinitionTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.expected = {"name": "Main", "type": "String"}

    def test_to_json(self):
        cpd = abi.ContractParameterDefinition("Main", abi.ContractParameterType.STRING)
        self.assertEqual(self.expected, cpd.to_json())

    def test_from_json(self):
        cpd = abi.ContractParameterDefinition.from_json(self.expected)
        self.assertEqual("Main", cpd.name)
        self.assertEqual(abi.ContractParameterType.STRING, cpd.type)

        with self.assertRaises(KeyError) as context:
            json_without_name = {"type": abi.ContractParameterType.ANY}
            abi.ContractParameterDefinition.from_json(json_without_name)
        self.assertIn("name", str(context.exception))

        with self.assertRaises(KeyError) as context:
            json_without_type = {"name": "Main"}
            abi.ContractParameterDefinition.from_json(json_without_type)
        self.assertIn("type", str(context.exception))

    def test_eq(self):
        cpd = abi.ContractParameterDefinition.from_json(self.expected)
        cpd2 = abi.ContractParameterDefinition.from_json(self.expected)
        self.assertFalse(cpd == object())
        self.assertTrue(cpd == cpd2)


class ContractEventDescriptorTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.expected = {
            "name": "MainEvent",
            "parameters": [{"name": "param", "type": "String"}],
        }
        cls.parameters = [
            abi.ContractParameterDefinition("param", abi.ContractParameterType.STRING)
        ]

    def test_to_json(self):
        ced = abi.ContractEventDescriptor("MainEvent", self.parameters)
        self.assertEqual(self.expected, ced.to_json())

    def test_from_json(self):
        ced = abi.ContractEventDescriptor.from_json(self.expected)
        self.assertEqual("MainEvent", ced.name)
        self.assertEqual(self.parameters, ced.parameters)

        with self.assertRaises(KeyError) as context:
            json_without_name = self.expected.copy()
            json_without_name.pop("name")
            abi.ContractEventDescriptor.from_json(json_without_name)
        self.assertIn("name", str(context.exception))

        with self.assertRaises(KeyError) as context:
            json_without_param = self.expected.copy()
            json_without_param.pop("parameters")
            abi.ContractEventDescriptor.from_json(json_without_param)
        self.assertIn("parameters", str(context.exception))

    def test_eq(self):
        ced = abi.ContractEventDescriptor.from_json(self.expected)
        ced2 = abi.ContractEventDescriptor.from_json(self.expected)
        self.assertFalse(ced == object())
        self.assertTrue(ced == ced2)


class ContractMethodDescriptorTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.parameters = [
            abi.ContractParameterDefinition("param", abi.ContractParameterType.STRING)
        ]
        cls.expected = {
            "name": "MainMethod",
            "offset": 0,
            "parameters": [cls.parameters[0].to_json()],
            "returntype": "Boolean",
            "safe": True,
        }

    def test_to_json(self):
        cmd = abi.ContractMethodDescriptor(
            "MainMethod", 0, self.parameters, abi.ContractParameterType.BOOLEAN, True
        )
        self.assertEqual(self.expected, cmd.to_json())

    def test_from_json(self):
        cmd = abi.ContractMethodDescriptor.from_json(self.expected)
        self.assertEqual("MainMethod", cmd.name)
        self.assertEqual(self.parameters, cmd.parameters)
        self.assertEqual(0, cmd.offset)
        self.assertEqual(abi.ContractParameterType.BOOLEAN, cmd.return_type)
        self.assertTrue(cmd.safe)

        with self.assertRaises(KeyError) as context:
            json_without_return_type = self.expected.copy()
            json_without_return_type.pop("returntype")
            abi.ContractMethodDescriptor.from_json(json_without_return_type)
        self.assertIn("returntype", str(context.exception))

    def test_eq(self):
        cmd = abi.ContractMethodDescriptor.from_json(self.expected)
        cmd2 = abi.ContractMethodDescriptor.from_json(self.expected)
        self.assertFalse(cmd == object())
        self.assertTrue(cmd == cmd2)


class AbiTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var methods = new ContractMethodDescriptor[]
        {
            new ContractMethodDescriptor {
                Name = "main_entry",
                Offset = 0,
                Parameters = new ContractParameterDefinition[0],
                ReturnType = ContractParameterType.Integer,
                Safe = true
            }
        };
        var events = new ContractEventDescriptor[]
        {
            new ContractEventDescriptor() { Name = "main_event", Parameters = new ContractParameterDefinition[0]}
        };
        var abi = new ContractAbi()
        {
            Methods = methods,
            Events = events
        };
        Console.WriteLine(abi.ToJson());
        """
        cls.method1 = abi.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=abi.ContractParameterType.INTEGER,
            safe=True,
        )
        cls.methods = [cls.method1]
        cls.event = abi.ContractEventDescriptor(name="main_event", parameters=[])
        cls.events = [cls.event]
        # captured from C#
        cls.expected_json = {
            "methods": [
                {
                    "name": "main_entry",
                    "parameters": [],
                    "returntype": "Integer",
                    "offset": 0,
                    "safe": True,
                }
            ],
            "events": [{"name": "main_event", "parameters": []}],
        }

    def test_to_json(self):
        abi_ = abi.ContractABI(methods=self.methods, events=self.events)

        self.assertEqual(self.expected_json, abi_.to_json())

    def test_from_json(self):
        # if "test_to_json" passes, then we know we can use our defined class
        # attributes to validate the from_json() results
        abi_ = abi.ContractABI.from_json(self.expected_json)
        self.assertEqual(self.methods, abi_.methods)
        self.assertEqual(self.events, abi_.events)

    def test_eq(self):
        abi_ = abi.ContractABI.from_json(self.expected_json)
        abi2_ = abi.ContractABI.from_json(self.expected_json)
        self.assertFalse(abi_ == object())
        self.assertTrue(abi_ == abi2_)

    def test_get_method(self):
        abi_ = abi.ContractABI.from_json(self.expected_json)
        self.assertIsNone(abi_.get_method("bad_method", 0))
        method = abi_.get_method("main_entry", 0)
        self.assertIsInstance(method, abi.ContractMethodDescriptor)
        self.assertEqual("main_entry", method.name)



"""
Contract utilities for determing the contract hash, contract types, extracting public keys and signing treshold and more.
"""
from typing import Type
from collections.abc import Sequence
from neo3.core import types, utils as coreutils, cryptography
from neo3 import vm


def get_contract_hash(
    sender: types.UInt160, nef_checksum: int, contract_name: str
) -> types.UInt160:
    """
    Return the calculated contract hash.

    Args:
        sender: script hash of account that deployed the contract.
        nef_checksum: checksum of the contract's NEF file.
        contract_name: the name from the contract's manifest.
    Returns:
        a unique contract identifier.
    """
    sb = vm.ScriptBuilder()
    sb.emit(vm.OpCode.ABORT)
    sb.emit_push(sender)
    sb.emit_push(nef_checksum)
    sb.emit_push(contract_name)
    return coreutils.to_script_hash(sb.to_array())


def validate_type(obj: object, type_: Type):
    """
    Helper function to validate ABI type information.

    Args:
        obj: target object.
        type_: expected type

    Raises:
        ValueError: if types do not match.
    """
    if type(obj) != type_:
        raise ValueError(f"Expected type '{type_}' , got '{type(obj)}' instead")
    return obj


def create_multisig_redeemscript(
    m: int, public_keys: Sequence[cryptography.ECPoint]
) -> bytes:
    """
    Create a multi-signature redeem script requiring `m` signatures from the list `public_keys`.

    This generated script is intended to be executed by the VM to indicate that the requested action is allowed.

    Args:
        m: minimum number of signature required for signing. Can't be lower than 2.
        public_keys: public keys to use during verification.

    Raises:
        ValueError: if the minimum required signatures is not met.
        ValueError: if the maximum allowed signatures is exceeded.
        ValueError: if the maximum allowed public keys is exceeded.
    """
    if m < 1:
        raise ValueError(f"Minimum required signature count is 1, specified {m}.")

    if m > len(public_keys):
        raise ValueError(
            "Invalid public key count. "
            "Minimum required signatures is bigger than supplied public keys count."
        )

    if len(public_keys) > 1024:
        raise ValueError(
            f"Supplied public key count ({len(public_keys)}) exceeds maximum of 1024."
        )

    sb = vm.ScriptBuilder()
    sb.emit_push(m)
    public_keys = list(public_keys)
    public_keys.sort()

    for key in public_keys:
        sb.emit_push(key.encode_point(True))

    sb.emit_push(len(public_keys))
    sb.emit_syscall(vm.Syscalls.SYSTEM_CRYPTO_CHECK_MULTI_SIGNATURE_ACCOUNT)
    return sb.to_array()


def create_signature_redeemscript(public_key: cryptography.ECPoint) -> bytes:
    """
    Create a single signature redeem script.

    This generated script is intended to be executed by the VM to indicate that the requested action is allowed.

    Args:
        public_key: the public key to use during verification.
    """
    sb = vm.ScriptBuilder()
    sb.emit_push(public_key.encode_point(True))
    sb.emit_syscall(vm.Syscalls.SYSTEM_CRYPTO_CHECK_STANDARD_ACCOUNT)
    return sb.to_array()


def is_signature_contract(script: bytes) -> bool:
    """
    Test if the provided script is a (single) signature contract.

    Args:
        script: contract script.
    """
    if len(script) != 40:
        return False

    if (
        script[0] != vm.OpCode.PUSHDATA1
        or script[1] != 33
        or script[35] != vm.OpCode.SYSCALL
        or script[36:40] != vm.Syscalls.SYSTEM_CRYPTO_CHECK_STANDARD_ACCOUNT
    ):
        return False
    return True


def is_multisig_contract(script: bytes) -> bool:
    """
    Test if the provided script is a multi-signature contract.

    Args:
        script: contract script.
    """
    valid, _, _ = parse_as_multisig_contract(script)
    return valid


def parse_as_multisig_contract(
    script: bytes,
) -> tuple[bool, int, list[cryptography.ECPoint]]:
    """
    Try to parse script as multisig contract and extract related data.

    Args:
        script: array of vm byte code.

    Returns:
        bool: `True` if the script passes as a valid multisignature contract script. `False` otherwise.
        int: the signing threshold if validation passed. 0 otherwise.
        list[ECPoint]: the public keys in the script if valiation passed. An empty array otherwise.
    """
    script = bytes(script)
    VALIDATION_FAILURE: tuple[bool, int, list[cryptography.ECPoint]] = (False, 0, [])

    len_script = len(script)
    if len_script < 42:
        return VALIDATION_FAILURE

    # read signature length, which is encoded as variable_length
    first_byte = script[0]
    if first_byte == int(vm.OpCode.PUSHINT8):
        signature_threshold = script[1]
        i = 2
    elif first_byte == int(vm.OpCode.PUSHINT16):
        signature_threshold = int.from_bytes(script[1:3], "little", signed=False)
        i = 3
    elif int(vm.OpCode.PUSH1) <= first_byte <= int(vm.OpCode.PUSH16):
        signature_threshold = first_byte - int(vm.OpCode.PUSH0)
        i = 1
    else:
        return VALIDATION_FAILURE

    if signature_threshold < 1 or signature_threshold > 1024:
        return VALIDATION_FAILURE

    # try reading public keys and do a basic format validation
    pushdata1 = int(vm.OpCode.PUSHDATA1)
    public_keys = []
    while script[i] == pushdata1:
        if len_script <= i + 35:
            return VALIDATION_FAILURE
        if script[i + 1] != 33:
            return VALIDATION_FAILURE
        public_keys.append(
            cryptography.ECPoint.deserialize_from_bytes(script[i + 2 : i + 2 + 33])
        )
        i += 35

    public_key_count = len(public_keys)
    if public_key_count < signature_threshold or public_key_count > 1024:
        return VALIDATION_FAILURE

    # validate that the number of collected public keys match the expected count
    value = script[i]
    if value == int(vm.OpCode.PUSHINT8):
        if len_script <= i + 1 or public_key_count != script[i + 1]:
            return VALIDATION_FAILURE
        i += 2
    elif value == int(vm.OpCode.PUSHINT16):
        if len_script < i + 3 or public_key_count != int.from_bytes(
            script[i + 1 : i + 3], "little", signed=False
        ):
            return VALIDATION_FAILURE
        i += 3
    elif int(vm.OpCode.PUSH1) <= value <= int(vm.OpCode.PUSH16):
        if public_key_count != value - int(vm.OpCode.PUSH0):
            return VALIDATION_FAILURE
        i += 1
    else:
        return VALIDATION_FAILURE

    if len_script != i + 5:
        return VALIDATION_FAILURE

    if script[i] != int(vm.OpCode.SYSCALL):
        return VALIDATION_FAILURE
    i += 1

    syscall_num = int.from_bytes(script[i : i + 4], "little")
    if syscall_num != vm.Syscalls.SYSTEM_CRYPTO_CHECK_MULTI_SIGNATURE_ACCOUNT:
        return VALIDATION_FAILURE
    return True, signature_threshold, public_keys



"""
P2P node. Connects over TCP. Handles network messages.
"""
from __future__ import annotations
import asyncio
import traceback
import struct
import string
from datetime import datetime
from neo3.network import message, capabilities, relaycache
from neo3.network.ipfilter import ipfilter
from neo3.network.convenience import nodeweight
from neo3.network.payloads import address, version, inventory, ping, block
from neo3 import network_logger as logger, settings
from neo3.core import types, msgrouter
from contextlib import suppress
from socket import AF_INET as IP4_FAMILY
from typing import Optional, Callable, cast
from asyncio.streams import StreamWriter, StreamReader
from neo3.network.message import Message
from collections.abc import Sequence


class NeoNode:
    """
    P2P network actor.
    """

    #: list[address.NetworkAddress]: a list of known network addresses (class attribute).
    addresses = []  # type: list[address.NetworkAddress]

    def __init__(self, reader: StreamReader, writer: StreamWriter):
        #: Unique identifier.
        self.nodeid: int = id(self)
        #: Address of the remote endpoint.
        self.address = address.NetworkAddress(
            "0.0.0.0:0", state=address.AddressState.DEAD
        )
        #: Human readable id.
        self.nodeid_human: str = encode_base62(self.nodeid)
        self.version: Optional[version.VersionPayload] = None
        self.tasks: list[asyncio.Task] = []
        self.nodeweight = nodeweight.NodeWeight(self.nodeid)
        #: Best block height of node.
        self.best_height: int = 0
        self.best_height_last_update = datetime.utcnow().timestamp()

        self._read_task: Optional[asyncio.Task] = None

        #: bool: Whether the node is in the process of disconnecting and shutting down its tasks.
        self.disconnecting: bool = False

        #: dict[message.MessageType, Callable[[message.Message], None]]: A table matching message types to handler
        #: functions.
        self.dispatch_table: dict[
            message.MessageType, Callable[[message.Message], None]
        ] = {
            message.MessageType.ADDR: self.handler_addr,
            message.MessageType.BLOCK: self.handler_block,
            message.MessageType.CONSENSUS: self.handler_consensus,
            message.MessageType.INV: self.handler_inv,
            message.MessageType.FILTERADD: self.handler_filteradd,
            message.MessageType.FILTERCLEAR: self.handler_filterclear,
            message.MessageType.FILTERLOAD: self.handler_filterload,
            message.MessageType.GETADDR: self.handler_getaddr,
            message.MessageType.GETBLOCKS: self.handler_getblocks,
            message.MessageType.GETBLOCKBYINDEX: self.handler_getblockdata,
            message.MessageType.GETDATA: self.handler_getdata,
            message.MessageType.GETHEADERS: self.handler_getheaders,
            message.MessageType.HEADERS: self.handler_headers,
            message.MessageType.MEMPOOL: self.handler_mempool,
            message.MessageType.MERKLEBLOCK: self.handler_merkleblock,
            message.MessageType.PING: self.handler_ping,
            message.MessageType.PONG: self.handler_pong,
            message.MessageType.TRANSACTION: self.handler_transaction,
            message.MessageType.EXTENSIBLE: self.handler_extensible,
        }

        self.reader = reader
        self.writer = writer

    def __eq__(self, other):
        if type(other) is type(self):
            return self.address == other.address and self.nodeid == other.nodeid
        else:
            return False

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))}> {self.nodeid_human}"

    def connection_made(self) -> None:
        """
        Event called by the NeoNode.connect_to.
        """
        addr_tuple = self.writer.get_extra_info("peername")
        addr = f"{addr_tuple[0]}:{addr_tuple[1]}"

        network_addr = self._find_address_by_host_port(addr)
        if network_addr:
            # this scenario occurs when the NodeManager queues seed nodes
            self.address = network_addr
        else:
            self.address.address = addr

        if not ipfilter.is_allowed(addr_tuple[0]):
            logger.debug(f"Blocked by ipfilter: {self.address.address}")
            self._create_task_with_cleanup(
                self.disconnect(address.DisconnectReason.IPFILTER_NOT_ALLOWED)
            )
            return

    async def disconnect(self, reason: address.DisconnectReason) -> None:
        """
        Close the connection to remote endpoint.

        Args:
            reason: reason for disconnecting.
        """
        if self.disconnecting:
            return

        self.disconnecting = True

        logger.debug(f"Disconnect called with reason={reason.name}")
        self.address.disconnect_reason = reason
        if reason in [
            address.DisconnectReason.MAX_CONNECTIONS_REACHED,
            address.DisconnectReason.POOR_PERFORMANCE,
            address.DisconnectReason.HANDSHAKE_VERACK_ERROR,
            address.DisconnectReason.HANDSHAKE_VERSION_ERROR,
            address.DisconnectReason.UNKNOWN,
        ]:
            self.address.set_state_poor()
        elif reason == address.DisconnectReason.IPFILTER_NOT_ALLOWED:
            self.address.set_state_dead()

        for t in self.tasks:
            t.cancel()
            with suppress(asyncio.CancelledError):
                logger.debug(f"waiting for task to cancel {t}.")
                await t
                logger.debug("done")
        msgrouter.on_node_disconnected(self, reason)
        self.writer.close()
        await self.writer.wait_closed()

    def handler_addr(self, msg: message.Message) -> None:
        """
        Handler for a message with the `ADDR` type.

        Args:
            msg:
        """
        payload = cast(address.AddrPayload, msg.payload)
        self.addresses = list(set(self.addresses + payload.addresses))
        msgrouter.on_addr(payload.addresses)

    def handler_block(self, msg: message.Message) -> None:
        """
        Handler for a message with the `BLOCK` type.

        Args:
            msg:
        """
        msgrouter.on_block(self.nodeid, msg.payload)

    def handler_consensus(self, msg: message.Message) -> None:
        """
        Handler for a message with the `CONSENSUS` type.

        Args:
            msg:
        """
        pass

    def handler_inv(self, msg: message.Message) -> None:
        """
        Handler for a message with the `INV` type.

        Args:
            msg:
        """
        payload = cast(inventory.InventoryPayload, msg.payload)
        if payload.type == inventory.InventoryType.BLOCK:
            # neo-cli broadcasts INV messages on a regular interval. We can use those as trigger to request
            # their latest block height
            if len(payload.hashes) > 0:
                height = 0
                m = message.Message(
                    msg_type=message.MessageType.PING,
                    payload=ping.PingPayload(height=height),
                )
                self._create_task_with_cleanup(self.send_message(m))
        else:
            logger.debug(
                f"Message with type INV received. No processing for payload type "  # type:ignore
                f"{payload.type.name} implemented"
            )

    def handler_filteradd(self, msg: message.Message) -> None:
        """
        Handler for a message with the `FILTERADD` type.

        Args:
            msg:
        """
        pass

    def handler_filterclear(self, msg: message.Message) -> None:
        """
        Handler for a message with the `FILTERCLEAR` type.

        Args:
            msg:
        """
        pass

    def handler_filterload(self, msg: message.Message) -> None:
        """
        Handler for a message with the `FILTERLOAD` type.

        Args:
            msg:
        """
        pass

    def handler_getaddr(self, msg: message.Message) -> None:
        """
        Handler for a message with the `GETADDR` type.

        Args:
            msg:
        """
        addr_list = []
        for addr in self.addresses:  # type: address.NetworkAddress
            if addr.is_state_new or addr.is_state_connected:
                addr_list.append(addr)
        self._create_task_with_cleanup(self.send_address_list(addr_list))

    def handler_getblocks(self, msg: message.Message) -> None:
        """
        Handler for a message with the `GETBLOCKS` type.

        Args:
            msg:
        """
        pass

    def handler_getblockdata(self, msg: message.Message) -> None:
        """
        Handler for a message with the `GETBLOCKBYINDEX` type.

        Args:
            msg:
        """
        pass

    def handler_getdata(self, msg: message.Message) -> None:
        """
        Handler for a message with the `GETDATA` type.

        Args:
            msg:
        """
        payload = cast(inventory.InventoryPayload, msg.payload)
        for h in payload.hashes:
            item = relaycache.RelayCache().try_get(h)
            if item is None:
                # for the time being we only support data retrieval for our own relays
                continue
            if payload.type == inventory.InventoryType.TX:
                m = message.Message(
                    msg_type=message.MessageType.TRANSACTION, payload=item
                )
                self._create_task_with_cleanup(self.send_message(m))

    def handler_getheaders(self, msg: message.Message) -> None:
        """
        Handler for a message with the `GETHEADERS` type.

        Args:
            msg:
        """
        pass

    def handler_mempool(self, msg: message.Message) -> None:
        """
        Handler for a message with the `MEMPOOL` type.

        Args:
            msg:
        """
        pass

    def handler_merkleblock(self, msg: message.Message) -> None:
        """
        Handler for a message with the `MERKLEBLOCK` type.

        Args:
            msg:
        """
        pass

    def handler_headers(self, msg: message.Message) -> None:
        """
        Handler for a message with the `HEADERS` type.

        Args:
            msg:
        """
        payload = cast(block.HeadersPayload, msg.payload)
        if len(payload.headers) > 0:
            msgrouter.on_headers(self.nodeid, payload.headers)

    def handler_ping(self, msg: message.Message) -> None:
        """
        Handler for a message with the `PING` type.

        Args:
            msg:
        """
        height = 0
        m = message.Message(
            msg_type=message.MessageType.PONG, payload=ping.PingPayload(height=height)
        )
        self._create_task_with_cleanup(self.send_message(m))

    def handler_pong(self, msg: message.Message) -> None:
        """
        Handler for a message with the `PONG` type.

        Args:
            msg:
        """
        payload = cast(ping.PingPayload, msg.payload)
        if self.best_height != payload.current_height:
            logger.debug(
                f"Updating node {self.nodeid_human} height "
                f"from {self.best_height} to {payload.current_height}"
            )
            self.best_height = payload.current_height
            self.best_height_last_update = datetime.utcnow().timestamp()

    def handler_transaction(self, msg: message.Message) -> None:
        """
        Handler for a message with the `TRANSACTION` type.

        Args:
            msg:
        """
        pass

    def handler_extensible(self, msg: message.Message) -> None:
        """
        Handler for a message with the `EXTENSIBLE` type.

        Args:
            msg:
        """
        pass

    def start_message_handler(self) -> None:
        """
        A convenience function to start a message reading loop and forward the messages to their respective handlers as
        configured in :attr:`~neo3.network.node.NeoNode.dispatch_table`.
        """
        # when we break out of the read/write loops, we should make sure we disconnect
        self._read_task = asyncio.create_task(self._process_incoming_data())
        self._read_task.add_done_callback(
            lambda _: asyncio.create_task(
                self.disconnect(address.DisconnectReason.UNKNOWN)
            )
        )

    async def send_message(self, msg: message.Message) -> None:
        """
        Broadcast a message to the network.
        """
        self.writer.write(msg.to_array())
        await self.writer.drain()

    async def read_message(self, timeout: Optional[int] = 30) -> Optional[Message]:
        """
        Read a message from the network.

        Args:
            timeout: maximum time to wait for a message to arrive over the network.
        """
        if timeout == 0:
            # avoid memleak. See: https://bugs.python.org/issue37042
            timeout = None

        async def _read():
            try:
                # readexactly can throw ConnectionResetError
                message_header = await self.reader.readexactly(3)
                payload_length = message_header[2]

                if payload_length == 0xFD:
                    len_bytes = await self.reader.readexactly(2)
                    (payload_length,) = struct.unpack("<H", len_bytes)
                elif payload_length == 0xFE:
                    len_bytes = await self.reader.readexactly(4)
                    (payload_length,) = struct.unpack("<I", len_bytes)
                elif payload_length == 0xFE:
                    len_bytes = await self.reader.readexactly(8)
                    (payload_length,) = struct.unpack("<Q", len_bytes)
                else:
                    len_bytes = b""

                if payload_length > Message.PAYLOAD_MAX_SIZE:
                    raise ValueError("Invalid format")

                payload_data = await self.reader.readexactly(payload_length)
                raw = message_header + len_bytes + payload_data

                try:
                    return Message.deserialize_from_bytes(raw)
                except Exception:
                    logger.debug(
                        f"Failed to deserialize message: {traceback.format_exc()}"
                    )
                    return None

            except (ConnectionResetError, ValueError) as e:
                # ensures we break out of the main run() loop of Node, which triggers a disconnect callback to clean up
                self.disconnecting = True
                logger.debug(
                    f"Failed to read message data for reason: {traceback.format_exc()}"
                )
                return None
            except (asyncio.CancelledError, asyncio.IncompleteReadError):
                return None
            except Exception:
                # ensures we break out of the main run() loop of Node, which triggers a disconnect callback to clean up
                logger.debug(f"error read message 1 {traceback.format_exc()}")
                return None

        try:
            return await asyncio.wait_for(_read(), timeout)
        except (asyncio.TimeoutError, asyncio.CancelledError):
            return None
        except Exception:
            logger.debug("error read message 2")
            traceback.print_exc()
            return None

    # raw network commands
    async def request_address_list(self) -> None:
        """
        Send a request for receiving known network addresses.
        """
        m = message.Message(msg_type=message.MessageType.GETADDR)
        await self.send_message(m)

    async def send_address_list(
        self, network_addresses: Sequence[address.NetworkAddress]
    ) -> None:
        """
        Send network addresses.

        Args:
            network_addresses: list of addresses of other network actors.
        """
        m = message.Message(
            msg_type=message.MessageType.ADDR,
            payload=address.AddrPayload(addresses=network_addresses),
        )
        await self.send_message(m)

    async def request_headers(
        self, index_start: int, count: int = block.HeadersPayload.MAX_HEADERS_COUNT
    ) -> None:
        """
        Send a request for headers from `index_start` to `index_start`+`count`.

        Not specifying a `count` results in requesting at most 2000 headers.

        Args:
            index_start: block height to start from.
            count: number of headers to request.
        """
        m = message.Message(
            msg_type=message.MessageType.GETHEADERS,
            payload=block.GetBlockByIndexPayload(index_start, count),
        )
        await self.send_message(m)

    async def send_headers(self, headers: Sequence[block.Header]) -> None:
        """
        Send a list of Header objects.
        """
        if len(headers) > 2000:
            headers = headers[:2000]

        m = message.Message(
            msg_type=message.MessageType.HEADERS, payload=block.HeadersPayload(headers)
        )
        await self.send_message(m)

    async def request_blocks(
        self, hash_start: types.UInt256, count: Optional[int] = None
    ) -> None:
        """
        Send a request for retrieving block hashes from `hash_start` to `hash_start`+`count`.

        Not specifying a `count` results in requesting at most 500 blocks.

        Note:
            The remote node is expected to reply with a Message with the :const:`~neo3.network.message.MessageType.INV`
            type containing the hashes of the requested blocks. Use :meth:`~neo3.network.node.NeoNode.request_data` in
            combination with these hashes to return the actual :class:`~neo3.network.payloads.block.Block` objects.

        See also:
            `NeoNode.request_block_data()` to immediately retrieve
            `neo3.network.payloads.block.Block` objects.

        Args:
            hash_start: block hash to start from.
            count: number of blocks to return.
        """
        m = message.Message(
            msg_type=message.MessageType.GETBLOCKS,
            payload=block.GetBlocksPayload(hash_start, count),
        )
        await self.send_message(m)

    async def request_block_data(self, index_start, count) -> None:
        """
        Send a request for `count` blocks starting from `index_start`.

        Count cannot exceed :attr:`~neo3.network.payloads.block.GetBlockByIndexPayload.MAX_BLOCKS_COUNT`.

        See also:
            :meth:`~neo3.network.node.NeoNode.request_blocks()` to only request block hashes.

        Args:
            index_start: block index to start from.
            count: number of blocks to return.
        """
        m = message.Message(
            msg_type=message.MessageType.GETBLOCKBYINDEX,
            payload=block.GetBlockByIndexPayload(index_start, count),
        )
        await self.send_message(m)

    async def request_data(
        self, type: inventory.InventoryType, hashes: Sequence[types.UInt256]
    ) -> None:
        """
        Send a request for receiving the specified inventory data.

        Args:
            type: the inventory type to request.
            hashes: the hashes of `type` to request.
        """
        if len(hashes) < 1:
            return

        m = message.Message(
            msg_type=message.MessageType.GETDATA,
            payload=inventory.InventoryPayload(type, hashes),
        )
        await self.send_message(m)

    async def send_inventory(
        self, inv_type: inventory.InventoryType, inv_hash: types.UInt256
    ) -> None:
        """
        Send an inventory to the network.

        Args:
            inv_type:
            inv_hash:
        """
        inv = inventory.InventoryPayload(type=inv_type, hashes=[inv_hash])
        m = message.Message(msg_type=message.MessageType.INV, payload=inv)
        await self.send_message(m)

    async def send_ping(self) -> None:
        """
        Send a Ping message and expecting a Pong response.
        """
        height = 0

        p = ping.PingPayload(height)
        m = message.Message(msg_type=message.MessageType.PING, payload=p)
        await self.send_message(m)

    async def relay(self, inv: inventory.IInventory) -> bool:
        """
        Relay the inventory to the network.

        Args:
            inv: should be of type Block, Transaction or Consensus.
        """
        relaycache.RelayCache().add(inv)
        await self.send_inventory(inv.inventory_type, inv.hash())
        return True

    @staticmethod
    async def connect_to(
        host: Optional[str] = None,
        port: Optional[int] = None,
        timeout=3,
        socket=None,
        *,
        _test_data: Optional[dict] = None,
    ) -> tuple[Optional[NeoNode], Optional[tuple[str, str]]]:
        """
        Connect to another node.

        Args:
            host: node ip.
            port: node port.
            timeout: max time to wait before aborting.
            socket: use an existing socket.
            _test_data:

        Returns:
            tuple[NeoNode, None]: if connection was established succesfully.
            tuple[None, tuple[host:port, failure_reason]]: if connection is not established successfully.
        """
        if host is not None or port is not None:
            if socket is not None:
                raise ValueError(
                    "host/port and socket can not be specified at the same time"
                )
        if socket is None and (host is None or port is None):
            raise ValueError("host and port was not specified and no sock specified")

        try:
            if socket:
                logger.debug(f"Trying to connect to socket: {socket}.")
                open_conn_coro = asyncio.open_connection(sock=socket)
            else:
                logger.debug(f"Trying to connect to: {host}:{port}.")
                open_conn_coro = asyncio.open_connection(host, port, family=IP4_FAMILY)
            reader, writer = await asyncio.wait_for(open_conn_coro, timeout)

            if _test_data and (peername := _test_data.get("peername")) is not None:
                writer._transport._extra["peername"] = peername  # type: ignore
            node = NeoNode(reader, writer)
            node.connection_made()

            success, fail_reason = await node._do_handshake()
            if success:
                return node, None
            else:
                raise Exception(fail_reason)
        except asyncio.TimeoutError:
            reason = "Timed out"
        except OSError as e:
            reason = f"Failed to connect for reason {e}"
        except asyncio.CancelledError:
            reason = "Cancelled"
        except Exception as e:
            reason = traceback.format_exc()
        return None, (f"{host}:{port}", reason)

    @classmethod
    def get_address_new(cls) -> Optional[address.NetworkAddress]:
        """
        Utility function to return the first address with the state NEW.
        """
        for addr in cls.addresses:
            if addr.is_state_new:
                return addr
        # explicit return to silence mypy
        return None

    async def _do_handshake(self) -> tuple[bool, Optional[address.DisconnectReason]]:
        caps: list[capabilities.NodeCapability] = [capabilities.FullNodeCapability(0)]
        # TODO: fix nonce and port if a service is running
        send_version = message.Message(
            msg_type=message.MessageType.VERSION,
            payload=version.VersionPayload(
                nonce=123, user_agent="NEO-MAMBA", capabilities=caps
            ),
        )
        await self.send_message(send_version)

        m = await self.read_message(timeout=3)
        if not m or m.type != message.MessageType.VERSION:
            await self.disconnect(address.DisconnectReason.HANDSHAKE_VERSION_ERROR)
            return False, address.DisconnectReason.HANDSHAKE_VERSION_ERROR

        if not self._validate_version(m.payload):
            await self.disconnect(address.DisconnectReason.HANDSHAKE_VERSION_ERROR)
            return False, address.DisconnectReason.HANDSHAKE_VERSION_ERROR

        m_verack = message.Message(msg_type=message.MessageType.VERACK)
        await self.send_message(m_verack)

        m = await self.read_message(timeout=3)
        if not m or m.type != message.MessageType.VERACK:
            await self.disconnect(address.DisconnectReason.HANDSHAKE_VERACK_ERROR)
            return False, address.DisconnectReason.HANDSHAKE_VERACK_ERROR

        user_agent = self.version.user_agent if self.version else ""
        logger.debug(
            f"Connected to {user_agent} @ {self.address.address}: {self.best_height}."
        )
        msgrouter.on_node_connected(self)

        return True, None

    def _create_task_with_cleanup(self, coro):
        task = asyncio.create_task(coro)
        self.tasks.append(task)
        task.add_done_callback(lambda fut: self.tasks.remove(fut))

    async def _process_incoming_data(self) -> None:
        """
        Main loop
        """
        logger.debug("Waiting for a message.")
        while not self.disconnecting:
            # we want to always listen for an incoming message
            m = await self.read_message(timeout=1)
            if m is None:
                continue

            handler = self.dispatch_table.get(m.type, None)
            if handler:
                handler(m)
            else:
                logger.debug(f"Unknown message with type: {m.type.name}.")

    def _find_address_by_host_port(self, host_port) -> Optional[address.NetworkAddress]:
        addr = address.NetworkAddress(address=host_port)
        try:
            idx = self.addresses.index(addr)
            return self.addresses[idx]
        except ValueError:
            return None

    def _validate_version(self, version) -> bool:
        if version.nonce == self.nodeid:
            logger.debug("Client is self.")
            return False

        if version.magic != settings.settings.network.magic:
            logger.debug(f"Wrong network id {version.magic}.")
            return False

        for c in version.capabilities:
            if isinstance(c, capabilities.ServerCapability):
                addr = self._find_address_by_host_port(self.address.address)

                if addr:
                    addr.set_state_connected()
                    addr.capabilities = version.capabilities
                else:
                    logger.debug(f"Adding address from outside {self.address.address}.")
                    # new connection initiated from outside
                    addr = address.NetworkAddress(
                        address=self.address.address,
                        capabilities=version.capabilities,
                        state=address.AddressState.CONNECTED,
                    )
                    self.addresses.append(addr)
                break

        for c in version.capabilities:
            if isinstance(c, capabilities.FullNodeCapability):
                # update nodes height indicator
                self.best_height = c.start_height
                self.best_height_last_update = datetime.utcnow().timestamp()
                self.version = version
                return True
        else:
            return False

    @classmethod
    def _reset_for_test(cls) -> None:
        cls.addresses = []


chars = string.digits + string.ascii_letters
base = len(chars)


def encode_base62(num: int):
    """Encode number in base62, returns a string."""
    if num < 0:
        raise ValueError("cannot encode negative numbers")

    if num == 0:
        return chars[0]

    digits = []
    while num:
        rem = num % base
        num = num // base
        digits.append(chars[rem])
    return "".join(reversed(digits))



import hashlib
from enum import Enum
from collections.abc import Sequence
from neo3.core import serialization, Size, types


def get_var_size(value: object) -> int:
    """
    Determine the variable size of an object.

    Note:
        This function is only used internally for correctly returning object sizes of network payloads to be in line
        with C#'s output.

    Args:
        value: any object type

    Raises:
        TypeError: if a specific Iterable type is not supported.
        ValueError: if a specific object type is not supported .
    """
    # public static int GetVarSize(this string value)
    if isinstance(value, str):
        value_size = len(value.encode("utf-8"))
        return get_var_size(value_size) + value_size

    # internal static int GetVarSize(int value)
    elif isinstance(value, int):
        if value < 0xFD:
            return Size.uint8
        elif value <= 0xFFFF:
            return Size.uint8 + Size.uint16
        else:
            return Size.uint8 + Size.uint32

    # internal static int GetVarSize<T>(this T[] value)
    elif isinstance(value, Sequence):
        value_length = len(value)
        value_size = 0

        if value_length > 0:
            if isinstance(value[0], serialization.ISerializable):
                value_size = sum(map(lambda t: len(t), value))
            elif isinstance(value[0], Enum):
                # Note: currently all Enum's in neo core (C#) are of type Byte. Only porting that part of the code
                value_size = value_length * Size.uint8
            elif isinstance(value, (bytes, bytearray)):
                # experimental replacement for: value_size = value.Length * Marshal.SizeOf<T>();
                # because I don't think we have a reliable 'SizeOf' in python
                value_size = value_length * Size.uint8
            else:
                raise TypeError(
                    f"Cannot accurately determine size of objects that do not inherit from 'ISerializable', "
                    f"'Enum' or 'bytes'. Found type: {type(value[0])}"
                )
    else:
        raise ValueError(
            f"[NOT SUPPORTED] Unexpected value type {type(value)} for get_var_size()"
        )

    return get_var_size(value_length) + value_size


def to_script_hash(data: bytes) -> types.UInt160:
    """
    Create a script hash based on the input data.

    Args:
        data: data to hash
    """
    intermediate_data = hashlib.sha256(data).digest()
    data_ = hashlib.new("ripemd160", intermediate_data).digest()
    return types.UInt160(data_)
